#include "gl_extensions.h"
#include "basic_shader.h"


static void enumerate(shader_program p);
static void execute(shader_program p);


basic_shader_t basic_shader = {

   INIT_MODULE_NAME("basic_shader")

   INIT_VERTEX_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "attribute vec3 av3vertex;"
      "attribute vec3 av3normal;"
      "attribute vec2 av2texture;"

      "uniform mat4 um44mvp;"
      "uniform mat4 um44mv;"
      "uniform vec3 uv3light;"
      "uniform vec4 uv4color;"
      "uniform float bias;"

      "varying vec4 vv4color;"
      "varying vec2 vv2texture;"

      "void main() {"
         "vec4 source = vec4(uv3light,0.0);"
         "vec4 n_norm = normalize(um44mv * vec4(av3normal,0.0));"
         "float test  = -dot(source,n_norm);"
         "if (test < bias)"
            "test = bias;"
         "float spot  = pow(10.0,test - 1.0);"
         "vv4color    = vec4(uv4color.xyz * clamp(spot,0.0,1.0),uv4color.w);"
         "vv2texture  = av2texture;"
         "gl_Position = um44mvp * vec4(av3vertex,1.0);"
      "}"

   INIT_FRAGMENT_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "uniform sampler2D texture1;"
      "uniform int have_texture;"

      "varying vec4 vv4color;"
      "varying vec2 vv2texture;"

      "void main(void) {"
         "vec4 color = vec4(1.0);"
         "if (have_texture > 0)"
            "color = texture2D(texture1, vv2texture);"
         "gl_FragColor = color * vv4color;"
      "}"

   INIT_ALLOCATED_SOURCES(0)
   INIT_STD_COMPILE_FUNCTION
   INIT_ENUMERATE_FUNCTION(&enumerate)
   INIT_EXECUTE_FUNCTION(&execute)
   INIT_SHADER_API_BLOCK_END
};

static GLint loc[10];

static void enumerate(shader_program p)
{
   basic_shader_t *s = (basic_shader_t*)p;

   if (!s || !glIsProgram(s->program))
      return;

   loc[0] = glGetUniformLocation(s->program,"um44mvp");
   loc[1] = glGetUniformLocation(s->program,"um44mv");
   loc[2] = glGetUniformLocation(s->program,"uv3light");
   loc[3] = glGetAttribLocation(s->program,"av3vertex");
   loc[4] = glGetAttribLocation(s->program,"av3normal");
   loc[5] = glGetAttribLocation(s->program,"av2texture");
   loc[6] = glGetUniformLocation(s->program,"uv4color");
   loc[7] = glGetUniformLocation(s->program,"texture1");
   loc[8] = glGetUniformLocation(s->program,"have_texture");
   loc[9] = glGetUniformLocation(s->program,"bias");


}

static void execute(shader_program p)
{
   basic_shader_t *s = (basic_shader_t*)p;

   if (!s)
      return;

   if (glIsProgram(s->program))
      glUseProgram(s->program);
   else
      return;

   glActiveTexture(GL_TEXTURE0);
   glUniform1i(loc[7],0);
   if (s->texture && glIsTexture(s->texture)) {
      glUniform1i(loc[8],1);
      glBindTexture(GL_TEXTURE_2D,s->texture);
   }
   else
      glUniform1i(loc[8],0);

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,s->vbo);


   glUniform1f(loc[9],s->bias);

   glUniformMatrix4fv(loc[0],1,GL_FALSE,(float*)s->um44mvp);
   glUniformMatrix4fv(loc[1],1,GL_FALSE,(float*)s->um44mv);
   glUniform3fv(loc[2],1,s->uv3light);
   glUniform4fv(loc[6],1,s->uv4color);

   glVertexAttribPointer(loc[3],3,GL_FLOAT,GL_FALSE,0,s->av3vertex);
   glEnableVertexAttribArray(loc[3]);
   glVertexAttribPointer(loc[4],3,GL_FLOAT,GL_FALSE,0,s->av3normal);
   glEnableVertexAttribArray(loc[4]);
   glVertexAttribPointer(loc[5],2,GL_FLOAT,GL_FALSE,0,s->av2texture);
   glEnableVertexAttribArray(loc[5]);

   glDrawArrays(GL_TRIANGLES,0,s->count);

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,0);
   glDisableVertexAttribArray(loc[3]);
   glDisableVertexAttribArray(loc[4]);
   glDisableVertexAttribArray(loc[5]);

   glUseProgram(0);
   glBindTexture(GL_TEXTURE_2D,0);
}
