#ifndef __CONFIG_LOADER_H__
#define __CONFIG_LOADER_H__

#ifdef __cplusplus
extern "C" {
#endif

typedef struct config_record config_record;
typedef struct config_loader config_loader;
typedef struct config_annotation config_annotation;

struct config_annotation {
	char 					*item;
	config_annotation	*next;
};

struct config_record {
	char 					*name;
	config_annotation	*doc;
	unsigned 			format;
	unsigned 			bytes;
	void 					*loc;		
	config_record 		*next;
};

struct config_loader {
	char					*name;
	config_annotation *doc;	
	config_record 		*list;
	config_loader		*link;
};

typedef struct {
	char					*name;
	config_annotation *doc;
	config_loader		*list;
}
config_aggregate;

/* Overlay which matches the front of all config element structures... */
typedef struct {
	char 					*name;
	config_annotation *doc;
}
config_item;
	

#define CONFIG_BYTE 	   1
#define CONFIG_INT 		2
#define CONFIG_UINT 		3
#define CONFIG_FLOAT 	4
#define CONFIG_DOUBLE 	5
#define CONFIG_STRING 	6


#define config_usei(ldr,var_name) (config_bind(ldr,#var_name,CONFIG_INT,sizeof(int),(void*)&(var_name)))
#define config_useu(ldr,var_name) (config_bind(ldr,#var_name,CONFIG_UINT,sizeof(int),(void*)&(var_name)))
#define config_usef(ldr,var_name) (config_bind(ldr,#var_name,CONFIG_FLOAT,sizeof(float),(void*)&(var_name)))
#define config_used(ldr,var_name) (config_bind(ldr,#var_name,CONFIG_DOUBLE,sizeof(double),(void*)&(var_name)))
#define config_useb(ldr,var_name) (config_bind(ldr,#var_name,CONFIG_BYTE,sizeof(char),(void*)&(var_name)))

#define config_usezs(ldr,var_name,sz) (config_bind(ldr,#var_name,CONFIG_STRING,sz,(void*)(var_name)))
#define config_useia(ldr,var_name,sz) (config_bind(ldr,#var_name,CONFIG_INT,(sz) * sizeof(int),(void*)(var_name)))
#define config_useua(ldr,var_name,sz) (config_bind(ldr,#var_name,CONFIG_UINT,(sz) * sizeof(int),(void*)(var_name)))
#define config_usefa(ldr,var_name,sz) (config_bind(ldr,#var_name,CONFIG_FLOAT,(sz) * sizeof(float),(void*)(var_name)))
#define config_useda(ldr,var_name,sz) (config_bind(ldr,#var_name,CONFIG_DOUBLE,(sz) * sizeof(double),(void*)(var_name)))
#define config_useba(ldr,var_name,sz) (config_bind(ldr,#var_name,CONFIG_BYTE,sz,(void*)(var_name)))


config_loader *create_config_loader(char *name);
void erase_config_loader(config_loader *cl);
void config_load(config_loader *cl, char *config_file);
void config_read(config_loader *cl, char *from_string);
config_record *config_bind(config_loader *cl, char *var_name, int type, 
																 int bytes, void *var_location);
void config_save(config_loader *cl, char *config_file);
char *config_to_new_string(config_loader *cl);

config_aggregate *create_config_aggregate(char *name);
void erase_config_aggregate(config_aggregate *ca);
void aggregate_register_config(config_aggregate *ca, config_loader *cl);
void aggregate_save_config(config_aggregate *ca, config_loader *cl);
void aggregate_load_config(config_aggregate *ca, config_loader *cl);

void config_doc(void *item,char *line_comment);


#ifdef __cplusplus
}
#endif

#endif /* __CONFIG_LOADER_H__ */

