#ifdef __cplusplus
extern "C" {
#endif

#define MAX_CREDS  100
#define CRED_TIME  7.0

#define USING_OPENGL
#include "platform.h"

#include "display.h"
#include "stage.h"
#include "ui_element.h"
#include "std_decor.h"
#include "renderer_api.h"
#include "aluf.h"
#include "file_parser.h"
#include "credits_scene.h"

#define CONFIG_ID "CREDITS_SCENE"

static const char *AUDIO_CREDITS_FILE = "./data/audio/credits.ogg";
static const char *CREDITS_FILE       = "./data/credits.txt";


typedef struct {

   IMPLEMENTS_SCENE_API

   ui_viewer    *viewer;
   ui_placard   *background;
   stream_src   *stream;
   ui_balloon   *title;
   ui_balloon   *img;
   ui_balloon   *text;
   float        time;

   ui_element  *hdr[MAX_CREDS];
   ui_element  *msg[MAX_CREDS];
   ui_element  *pic[MAX_CREDS];
   unsigned    cred_cnt;
   unsigned    cred_at;
   unsigned    lock;
}
credits_scene;

static void init(scene_api *object);
static void update(scene_api *object);
static void draw(scene_api *object);
static void enter(scene_api *object);
static void leave(scene_api *object);
static void erase(scene_api *object);
static void audio_level(scene_api *object,float lvl);
static void on_grow(ui_balloon *b, ui_update *u);
static void on_shrink(ui_balloon *b, ui_update *u);
static void reset_balloon(ui_balloon *b);
static int compile_credits(parse_list *p,credits_scene *c);

static credits_scene credits_scene_proto = {
   &init,
   &update,
   &draw,
   &enter,
   &leave,
   &erase,
   &audio_level,
   .7,
   0,0.001800,30.0,0.05,
   0,0,0,
   0,0,0,
   0,0,0,
   0,

   0,0,0,0
};

static color_t clr1 = {.8,.8,.8,.8};
static color_t clr2 = {.8,.8,.8,.8};

scene_api *create_credits_scene(config_aggregate *ca)
{
   credits_scene *c = (credits_scene*)malloc(sizeof(credits_scene));

   ui_placard *background;
   parse_list *pl;

   if (c) {
      /* c->cl = create_config_loader(CONFIG_ID);
         aggregate_register_config(ca,c->cl); */

      *c = credits_scene_proto;

      c->context = renderer_generate_unique_id();
      renderer_set_context_group(c->context);

      c->viewer = create_viewer(frame_state.overlay_area);
      c->viewer->pointer = 0;

      background = create_placard();
      background->look = material_is(
         .5,.5,.5,.5,
         renderer_register_texture_file("./data/images/maps/DSS2.jpg"));

      c->viewer->screen->add_item(c->viewer->screen,background);
      c->background = background;

      pl = file_parser_get_objects(CREDITS_FILE,P_FILE);

      if (pl && compile_credits(pl,c)) {

         c->title = create_balloon();
         c->title->expand_down = 1;
         c->title->id = 1;
         c->title->owner = (void*)c;
         c->title->on_grow = &on_grow;
         c->title->on_shrink = &on_shrink;
         c->title->add_item(c->title,c->hdr[c->cred_at]);

         c->img = create_balloon();
         c->img->id = 2;
         c->img->owner = (void*)c;
         c->img->on_grow = &on_grow;
         c->img->on_shrink = &on_shrink;
         c->img->add_item(c->img,c->pic[c->cred_at]);

         c->text = create_balloon();
         c->text->expand_left = 1;
         c->text->id = 3;
         c->text->owner = (void*)c;
         c->text->on_grow = &on_grow;
         c->text->on_shrink = &on_shrink;
         c->text->add_item(c->text,c->msg[c->cred_at]);

         c->viewer->screen->add_items(
            c->viewer->screen,3,
            c->title,
            c->img,
            c->text);

         file_parser_erase_list(pl);
      }
   }

   return (scene_api*)c;
}

static int compile_credits(parse_list *pl, credits_scene *c)
{
   unsigned id = 0;

   parse_obj *obj;
   ui_element *e;
   ui_text *tx;

   if (!c)
      return 0;

   while (pl) {

      obj = pl->p;
      if (!obj) {
         id++;
         pl = pl->more;
         continue;
      }

      if (obj->hdr) {
         tx = create_text(40,obj->hdr);
         tx->font = ui_pretty_fnt;
         tx->centered = 1;
         e = to_static_text(tx);
         e->place(e,0,.8,1,1);
         tx = 0;
         obj->hdr = 0;

         if (obj->hdr_clr.a) {
            decor_all_colors(
               (ui_decor*)&e->look,
               obj->hdr_clr.r,obj->hdr_clr.g,obj->hdr_clr.b,obj->hdr_clr.a);
         }
         else {
            decor_all_colors(
               (ui_decor*)&e->look,
               clr2.r,clr2.g,clr2.b,clr2.a);
         }

         c->hdr[id] = e;
         e = 0;
      }
      else c->hdr[id] = 0;

      if (obj->msg) {
         tx = create_text(80,obj->msg);
         tx->font = ui_pretty_fnt;
         tx->centered = 1;
         e = to_static_text(tx);
         e->place(e,0,.05,1,1);
         tx = 0;
         obj->msg = 0;

         if (obj->msg_clr.a) {
            decor_all_colors(
               (ui_decor*)&e->look,
               obj->msg_clr.r,obj->msg_clr.g,obj->msg_clr.b,obj->msg_clr.a);
         }
         else {
            decor_all_colors(
               (ui_decor*)&e->look,
               clr1.r,clr1.g,clr1.b,clr1.a);
         }

         c->msg[id] = e;
         e = 0;
      }
      else c->msg[id] = 0;

      if (obj->pic) {
         e = (ui_element*)create_custom_icon(.1);
         decor_all_materials(
            (ui_decor*)&e->look,
            .9,.9,.9,.9,
            renderer_register_texture_file(obj->pic));
         e->place(e,.3,.3,.4,.4);
         c->pic[id] = e;
         e = 0;
      }
      else c->pic[id] = 0;

      id++;
      pl = pl->more;
   }

   c->cred_cnt = id;
   return 1;
}

static void on_grow(ui_balloon *b, ui_update *u)
{
   if (!b)
      return;

   b->expand_left = !b->expand_left;
}

static void on_shrink(ui_balloon *b, ui_update *u)
{
   credits_scene *c;

   if (!b)
      return;

   c = (credits_scene*)b->owner;

   b->expand_down = !b->expand_down;
   b->detach_all(b);

   switch (b->id) {
      case 1:
         b->add_item(b,c->hdr[c->cred_at]);
         break;
      case 2:
         b->add_item(b,c->pic[c->cred_at]);
         break;
      case 3:
         b->add_item(b,c->msg[c->cred_at]);
         break;
      default:
         break;
    }
}

static void reset_balloon(ui_balloon *b)
{
   b->at = in_space(0,0,0,0);
   b->change = 0;
   b->presence = 0;
}

static void erase(scene_api *object)
{
   credits_scene *c = (credits_scene*)object;
   unsigned n = 0;

   if (!c)
      return;

   c->title->detach_all(c->title);
   c->img->detach_all(c->img);
   c->text->detach_all(c->text);

   while (n < c->cred_cnt) {
      if (c->hdr[n])
         c->hdr[n]->erase(c->hdr[n]);
      if (c->pic[n])
         c->pic[n]->erase(c->pic[n]);
      if (c->msg[n])
         c->msg[n]->erase(c->msg[n]);
      n++;
   }

   if (c->viewer)
      c->viewer->erase(c->viewer);

   if (c->stream)
      erase_stream_source(c->stream);

   free(c);
}

static void init(scene_api *object)
{
   credits_scene *c = (credits_scene*)object;

    if (!c)
      return;
}

static void update(scene_api *object)
{
   credits_scene *c = (credits_scene*)object;

   if (!c)
      return;

   c->time += frame_state.dt;
   if (!c->lock &&c->time > CRED_TIME) {
      c->cred_at++;
      if (c->cred_at == c->cred_cnt)
         c->cred_at = 0;
      c->title->shrink(c->title);
      c->img->shrink(c->img);
      c->text->shrink(c->text);
      c->lock = 1;
   }
   if (c->time > CRED_TIME + .5) {
      c->title->grow(c->title);
      c->img->grow(c->img);
      c->text->grow(c->text);
      c->time = 0;
      c->lock = 0;
   }

   c->viewer->update(
      c->viewer,
      frame_state.pointer[0],
      frame_state.pointer[1],
      frame_state.dt,
      frame_state.button,
      frame_state.key);

   if (ctl_button(4,frame_state.button))
      stage_pop_scene();
}

static void draw(scene_api *object)
{
   credits_scene *c = (credits_scene*)object;

   if (!c)
      return;

   scene_api_gl_defaults();

   c->viewer->display(c->viewer, c->projection);
}

static void enter(scene_api *object)
{
   credits_scene *c = (credits_scene*)object;

   if (!c)
      return;

   if (!c->stream)
      c->stream = create_stream_source(AUDIO_CREDITS_FILE);

   if (c->stream)
      alSourcePlay(c->stream->src);

   c->time = 0;
   c->title->grow(c->title);
   c->img->grow(c->img);
   c->text->grow(c->text);
}

static void leave(scene_api *object)
{
   credits_scene *c = (credits_scene*)object;

   if (!c)
      return;

   if (c->stream) {
      erase_stream_source(c->stream);
      c->stream = 0;
   }

   reset_balloon(c->title);
   reset_balloon(c->img);
   reset_balloon(c->text);
}

static void audio_level(scene_api *object,float lvl)
{
   credits_scene *c = (credits_scene*)object;

   if (c->stream) {
      alSourcef(c->stream->src,AL_GAIN,lvl);
   }
}

#ifdef __cplusplus
}
#endif
