#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include "display.h"
#include "stage.h"
#include "std_decor.h"
#include "aluf.h"
#include "display_controls.h"


#define USING_OPENAL_RUNTIME_LD
#include "ext_libs.h"

enum {ctl_clip, ctl_sep, ctl_focal, ctl_done, mode_ctrl, ctl_audio, exit_menu,
      ctl_fps, ctl_save};

typedef struct {
   IMPLEMENTS_UI_ELEMENT
   AND_UI_CONTAINER
   AND_UI_BALLOON

   ui_container   *items;
   ALuint         audio_chg;

   void (*save_erase)(void *obj);
}
popup_menu;

static void switchboard(void *obj,ui_update *u);
static void erase_stage_menu(void *obj);
static void fps_updater(ui_text_value *tx);

static unsigned use_fps = 0;

ui_balloon *create_display_common_controls(scene_api *sc)
{
   popup_menu *m = (popup_menu*)calloc(1,sizeof(popup_menu));
   ui_balloon *b;
   ui_icon *i;
   ui_button *bt;
   ui_container *c;
   ui_slider *s;
   float temp;

   if (!m)
      return 0;

   b = create_balloon();

   if (!b) {
      free(m);
      return 0;
   }

   memcpy(m,b,sizeof(ui_balloon));
   free(b);
   b = (ui_balloon*)m;

   b->owner = (void*)sc;

   i = create_custom_icon(.1);
   i->place(i,0,0,1,.5);
   i->look = material_is(0,0,0,.85,0);
   b->add_item(b,i);

   bt = create_button();
   bt->place(bt,.05,.05,.1,.1);
   *((ui_decor*)&bt->look) = std_rb_decor;
   bt->id = ctl_done;
   bt->on_use = &switchboard;
   bt->owner = (void*)m;
   b->add_item(b,bt);
   b->add_item(b,create_item_label(bt,"Done"));

   bt = create_button();
   bt->place(bt,.05,.2,.1,.1);
   *((ui_decor*)&bt->look) = std_rb_decor;
   bt->id = ctl_save;
   bt->on_use = &switchboard;
   bt->owner = (void*)sc;
   b->add_item(b,bt);
   b->add_item(b,create_item_label(bt,"Save"));

   c = create_container();
   c->place(c,.2,.2,1,1);

   if (display_set_clip_plane) {

      s = create_slider();
      s->place(s,0,0,.6,.05);
      *((ui_decor*)&s->look) = std_sl_decor;
      s->handle_width = .065;
      s->owner  = (void*)sc;
      s->id     = ctl_clip;
      s->min    = .0001;
      s->span   = 10;
      s->value  = sc->clipping_plane;
      s->on_use = &switchboard;
      c->add_item(c,create_slider_labels(s,3,"0","5","10"));
      c->add_item(c,create_item_label(s,"Front Clip"));
      c->add_item(c,s);
   }

   if (display_set_eye_seperation) {

      s = create_slider();
      s->place(s,0,.1,.6,.05);
      *((ui_decor*)&s->look) = std_sl_decor;
      s->handle_width = .065;
      s->owner  = (void*)sc;
      s->id     = ctl_sep;
      s->min    = -1;
      s->span   = 2;
      s->detents = 1;
      s->detent_width = .01;

      /* Covert "real" value back to normalized 3 * cube value */
      s->value = pow(fabs(sc->eye_seperation/3),1/3.0);
      if (sc->eye_seperation < 0)
         s->value = -s->value;

      s->on_use = &switchboard;
      c->add_item(c,create_slider_labels(s,7,
         "","More (Reversed)","","None","","More",""));
      c->add_item(c,create_item_label(s,"Eye Separation"));
      c->add_item(c,s);
   }

   if (display_set_focal_plane) {

      s = create_slider();
      s->place(s,0,.2,.6,.05);
      *((ui_decor*)&s->look) = std_sl_decor;
      s->handle_width = .065;
      s->owner  = (void*)sc;
      s->id     = ctl_focal;
      s->min    = 2;
      s->span   = 98;
      s->value  = sc->focal_plane;
      s->on_use = &switchboard;
      c->add_item(c,create_slider_labels(s,3,"2","51","100"));
      c->add_item(c,create_item_label(s,"Focal Plane"));
      c->add_item(c,s);
   }

   b->add_item(b,c);

   return (ui_balloon*)m;
}

ui_balloon *create_display_stage_controls(void)
{
   popup_menu *m = (popup_menu*)calloc(1,sizeof(popup_menu));
   ui_balloon *b;
   ui_container *c;
   ui_element *e;
   ui_icon *i;
   ui_button *bt;
   ui_slider *sl;
   ui_text *tx;
   ALuint buffer;
   int n = 0, *chain;
   float pad;

   if (!m)
      return 0;

   b = create_balloon();

   if (!b) {
      free(m);
      return 0;
   }

   memcpy(m,b,sizeof(ui_balloon));
   free(b);
   b = (ui_balloon*)m;

   m->save_erase = b->erase;
   b->erase = &erase_stage_menu;

   b->expand_left = 1;
   b->expand_down = 1;
   b->button_mask = 48;
   b->owner = (void*)b;
   b->id = exit_menu;
   b->on_use = &switchboard;

   /* The options pop up layout... */
   i = create_custom_icon(.1);
   i->place(i,0,.1,1,.5);
   i->look = material_is(0,0,0,.9,0);
   b->add_item(b,i);

    /* FPS button... */
   tx = create_text(80,"Information");
   tx->font = ui_pretty_bld_fnt;
   e = to_static_text(tx);
   e->place(e,.765,.315,1,1);
   decor_all_colors((ui_decor*)&e->look,.2,.5,.2,.55);
   b->add_item(b,e);

   bt = create_button();
   bt->place(bt,.8,.165,.1,.1);
   *((ui_decor*)&bt->look) = std_gr_gn_decor;
   bt->id = ctl_fps;
   bt->value = use_fps;
   bt->latches = 1;
   bt->owner  = (void*)m;
   bt->on_use = &switchboard;
   b->add_item(b,bt);
   b->add_item(b,create_item_label(bt,"FPS"));

   tx = create_text(80,"Audio");
   tx->font = ui_pretty_bld_fnt;
   e = to_static_text(tx);
   e->place(e,.075,.315,1,1);
   decor_all_colors((ui_decor*)&e->look,.2,.5,.2,.55);
   b->add_item(b,e);

   sl = create_slider();
   sl->place(sl,.09,.2,.6,.05);
   *((ui_decor*)&sl->look) = std_sl_decor;
   sl->handle_width = .065;
   sl->owner  = (void*)m;
   sl->id     = ctl_audio;
   sl->min    = 0;
   sl->span   = 1;
   sl->value  = stage_get_master_audio_level();
   sl->on_use = &switchboard;
   b->add_item(b,create_item_label(sl,"Level"));
   b->add_item(b,create_slider_labels(sl,5,"0","25","50","75","100"));
   b->add_item(b,sl);

   if (frame_state.stereo_modes > 0) {

      c = create_container();
      c->place(c,0,.385,1,1);

      tx = create_text(80,"3D Modes");
      tx->font = ui_pretty_bld_fnt;

      e = to_static_text(tx);
      e->place(e,.075,.165,1,1);
      decor_all_colors((ui_decor*)&e->look,.2,.5,.2,.55);
      c->add_item(c,e);

      pad = .45/frame_state.stereo_modes;
      while (n <= frame_state.stereo_modes) {

         bt = create_button();
         bt->place(bt,pad + ((float)n)/(frame_state.stereo_modes + 2),0,.1,.1);
         *((ui_decor*)&bt->look) = std_gr_gn_decor;

         if (!n) {
            chain = &bt->value;
            bt->value = frame_state.current_mode;
         }
         bt->on_use = &switchboard;
         bt->chain = chain;
         bt->value_down = n;
         bt->id = mode_ctrl;
         c->add_item(c,bt);
         c->add_item(c,create_item_label(bt,
            frame_state.stereo_mode_names[n]));

         n++;
      }

      b->add_item(b,c);
   }

   alGenSources(1,&m->audio_chg);
   buffer = aluf_buffer_from("./data/audio/cameraclick.wav");
   alSourcei(m->audio_chg,AL_BUFFER,buffer);

   return (ui_balloon*)m;
}

ui_element *retrieve_common_display(void)
{
   ui_text_value *tx;
   ui_container *c;

   if (!use_fps)
      return 0;

   c = create_container();
   tx = create_text_value(10,20,&fps_updater);
   tx->font = ui_pretty_fnt;
   tx->place(tx,.9,.975,.1,.1);
   decor_all_colors((ui_decor*)&tx->look,0,.4,0,.8);
   c->add_item(c,tx);

   return (ui_element*)c;
}

static void fps_updater(ui_text_value *tx)
{
   if (!tx)
      return;

   snprintf(tx->text,tx->length,"%4d",frame_state.fps);
}


static void erase_stage_menu(void *obj)
{
   popup_menu *m = (popup_menu*)obj;
   ALuint buffer;

   if (!m)
      return;

   alGetSourcei(m->audio_chg,AL_BUFFER,(ALint*)&buffer);
   if (buffer) {
      alSourcei(m->audio_chg,AL_BUFFER,0);
      alDeleteBuffers(1,&buffer);
   }
   alDeleteSources(1,&m->audio_chg);

   m->save_erase(m);
}


static void switchboard(void *obj,ui_update *u)
{
   ui_element *e = (ui_element*)obj;
   popup_menu *m;
   ui_balloon *b;
   scene_api *s;
   float cubic_curve;

   if (!e)
      return;

   s = (scene_api*)e->owner;

   switch (e->id) {
      case ctl_done:
         b = (ui_balloon*)e->owner;
         b->shrink(b);
         break;
      case exit_menu:
         b = (ui_balloon*)e;
         if (ctl_button(4,u->button))
            b->shrink(b);
         break;
      case ctl_clip:
         s = (scene_api*)e->owner;
         s->clipping_plane = ((ui_slider*)e)->value;
         display_set_clip_plane(s->clipping_plane);
         break;
      case ctl_sep:
         s = (scene_api*)e->owner;
         cubic_curve = ((ui_slider*)e)->value;
         cubic_curve = cubic_curve * cubic_curve * cubic_curve;
         s->eye_seperation = 3 * cubic_curve;
         display_set_eye_seperation(s->eye_seperation);
         break;
      case ctl_focal:
         s = (scene_api*)e->owner;
         s->focal_plane = ((ui_slider*)e)->value;
         display_set_focal_plane(s->focal_plane);
         break;
      case mode_ctrl:
         display_set_stereo_mode(*(((ui_button*)e)->chain));
         break;
      case ctl_audio:
         m = (popup_menu*)e->owner;
         stage_set_master_audio_level(((ui_slider*)e)->value);
         alSourcef(m->audio_chg,AL_GAIN,((ui_slider*)e)->value);
         if (!e->busy)
            alSourcePlay(m->audio_chg);
         break;
      case ctl_fps:
         use_fps = ((ui_button*)e)->value;
         break;
      case ctl_save:
         s = (scene_api*)e->owner;
         if (s && s->ca && s->cl)
            aggregate_save_config(s->ca,s->cl);
         break;
   }
}


#ifdef __cplusplus
}
#endif
