#ifndef __FILE_PARSER_H__
#define __FILE_PARSER_H__

#ifdef __cplusplus
extern "C" {
#endif

/* Notes:

The parsable file format will be as follows:

   document          = comment* def+ comment*
   comment           = '#' ascii_chars* '\n'
   def               = name body
   name              = character+ | nil

   body              = '{' comment* at property* '}'
   property          = comment* | clr | hdr | msg | nar | pic

   at                = 'at' '{' af ',' bf ',' cf ',' df '}'
   clr               = clr_def | nil
   hdr               = hdr_def | nil
   msg               = msg_def | nil
   nar               = nar_def | nil
   pic               = pic_def | nil

   clr_def           = 'clr' '{' rf ',' gf ',' bf ',' af '}'
   hdr_def           = 'hdr' '{' text_line* '}'
   msg_def           = 'msg' '{' text_line* '}'
   nar_def           = 'nar' '{' text_line '}'
   pic_def           = 'pic' '{' text_line '}'

   text_line         = character* comment*
   character         = non_reserved_char | escaped_char

   reserved_chars    = { '{' , '}' , '#' , '\' }
   ascii_chars       = ' ' through '~' (32 to 126 inclusive)
   non_reserved_char = ascii_chars - reserved_chars
   escaped_char      = '\' + reserved_char

Example document:

# Comments as such..
# Use this form multi-line...

name {
   at {x,y,z,r}  #allow annotation comment...
   clr {r,g,b,a} #allow comments here, too...
   msg {

      Whatever we want to type..

      Using escaped characters :
      \{ You are \#1 ! \}

  }
}

name2 {
   at {x1,y1,z1,r1}
   clr {r1,g1,b1,a1}
   msg {

      Whatever describes this...
   }
}


etc... */

typedef struct { float r, g, b, a; }      color_t;
typedef struct { float x, y, z, r; }      locat_t;
typedef struct { float x, y, w, h; }      area_t;
typedef union  { locat_t v; area_t a; }   at_t;

typedef struct loc_list loc_list;
struct loc_list {
   loc_list *more;
   at_t at;
};

typedef struct  {
   loc_list *at;

   color_t hdr_clr;
   char *hdr;

   color_t msg_clr;
   char *msg;

   char *nar;
   char *pic;
}
parse_obj;

typedef struct parse_list parse_list;
struct parse_list {
   parse_list *more;
   parse_obj *p;
};

enum { P_FILE = 1, P_STRING } parse_mode;

parse_list *file_parser_get_objects(const char* input, unsigned mode);
void file_parser_erase_list(parse_list *list);


#ifdef __cplusplus
}
#endif

#endif /* __FILE_PARSER_H__ */