#ifndef __GEOM_OBJECT_H__
#define __GEOM_OBJECT_H__
#include "geom_list.h"

typedef struct {
   void *array;         /* Data packed tight end-to-end, suitable for direct
                        transfer to VBO. Offsets are absolute with respect
                        to bytes from the begining of the array.
                        A negative value of the count/offset members denotes
                        that the item is not present in the array.
                        Order of occurance(if present) is the same as order
                        of declaration, i.e. vertices, normals, texture
                        coordinates, colors. */
   unsigned array_bytes;
   int vertex_count;    /* 3 * sizeof(float) per vertex coordinate. */
   int normal_offset;   /* 3 * sizeof(float) per normal vector. */
   int texture_offset;  /* 2 * sizeof(float) per texture coordinate. */
   int color_offset;    /* 4 * sizeof(float) per color vertex. */

}
geom_object;



geom_object *geom_object_create_sphere(int divisions, int sectors, unsigned inside);
geom_object *geom_object_create_washer(int divisions, int sectors, float center_coef);
geom_object *geom_object_create_basic_cube(void);
geom_object *geom_object_create_billboard(float aspect);
void geom_object_draw(geom_object *g);
void geom_object_free(geom_object *g);

/* Frees the parameter geom_object and returns a new geom_list */
geom_list *convert_to_geom_list(geom_object *g);


#endif /* __GEOM_OBJECT_H__ */