#include "gl_extensions.h"

#ifdef __cplusplus
"C" {
#endif

PFNGLGENBUFFERSPROC                    glGenBuffers;
PFNGLBINDBUFFERPROC                    glBindBuffer;
PFNGLBUFFERDATAPROC                    glBufferData;
PFNGLBUFFERSUBDATAPROC                 glBufferSubData;
PFNGLDELETEBUFFERSPROC                 glDeleteBuffers;
PFNGLGETBUFFERPARAMETERIVPROC          glGetBufferParameteriv;
PFNGLMAPBUFFERPROC                     glMapBuffer;
PFNGLUNMAPBUFFERPROC                   glUnmapBuffer;
PFNGLDRAWRANGEELEMENTSPROC             glDrawRangeElements;
PFNGLGETBUFFERSUBDATAPROC              glGetBufferSubData;
PFNGLISBUFFERPROC                      glIsBuffer;

PFNGLCREATEPROGRAMPROC                 glCreateProgram;
PFNGLDELETEPROGRAMPROC                 glDeleteProgram;
PFNGLISPROGRAMPROC                     glIsProgram;
PFNGLLINKPROGRAMPROC                   glLinkProgram;
PFNGLUSEPROGRAMPROC                    glUseProgram;
PFNGLATTACHSHADERPROC                  glAttachShader;
PFNGLDETACHSHADERPROC                  glDetachShader;
PFNGLGETPROGRAMIVPROC                  glGetProgramiv;
PFNGLGETPROGRAMINFOLOGPROC             glGetProgramInfoLog;
PFNGLVALIDATEPROGRAMPROC               glValidateProgram;

PFNGLCREATESHADERPROC                  glCreateShader;
PFNGLDELETESHADERPROC                  glDeleteShader;
PFNGLISSHADERPROC                      glIsShader;
PFNGLSHADERSOURCEPROC                  glShaderSource;
PFNGLCOMPILESHADERPROC                 glCompileShader;
PFNGLGETSHADERIVPROC                   glGetShaderiv;
PFNGLGETSHADERINFOLOGPROC              glGetShaderInfoLog;

PFNGLGETATTRIBLOCATIONPROC             glGetAttribLocation;
PFNGLBINDATTRIBLOCATIONPROC            glBindAttribLocation;
PFNGLVERTEXATTRIBPOINTERPROC           glVertexAttribPointer;
PFNGLENABLEVERTEXATTRIBARRAYPROC       glEnableVertexAttribArray;
PFNGLDISABLEVERTEXATTRIBARRAYPROC      glDisableVertexAttribArray;

PFNGLGETUNIFORMLOCATIONPROC            glGetUniformLocation;

PFNGLUNIFORM1FPROC                     glUniform1f;
PFNGLUNIFORM2FPROC                     glUniform2f;
PFNGLUNIFORM3FPROC                     glUniform3f;
PFNGLUNIFORM4FPROC                     glUniform4f;

PFNGLUNIFORM1IPROC                     glUniform1i;
PFNGLUNIFORM2IPROC                     glUniform2i;
PFNGLUNIFORM3IPROC                     glUniform3i;
PFNGLUNIFORM4IPROC                     glUniform4i;

PFNGLUNIFORM1FVPROC                    glUniform1fv;
PFNGLUNIFORM2FVPROC                    glUniform2fv;
PFNGLUNIFORM3FVPROC                    glUniform3fv;
PFNGLUNIFORM4FVPROC                    glUniform4fv;
PFNGLUNIFORM1IVPROC                    glUniform1iv;
PFNGLUNIFORM2IVPROC                    glUniform2iv;
PFNGLUNIFORM3IVPROC                    glUniform3iv;
PFNGLUNIFORM4IVPROC                    glUniform4iv;

PFNGLUNIFORMMATRIX2FVPROC              glUniformMatrix2fv;
PFNGLUNIFORMMATRIX3FVPROC              glUniformMatrix3fv;
PFNGLUNIFORMMATRIX4FVPROC              glUniformMatrix4fv;

PFNGLACTIVETEXTUREPROC                glActiveTexture;


static initialized = 0;


static unsigned populate_functions(void);

unsigned gl_extensions_init(void)
{
   if (initialized)
      return 1;

   initialized = populate_functions();
   return initialized;
}

static unsigned populate_functions(void)
{
   unsigned ok = 1;

   #define GB_GL(t,n)\
     ok &= (n = (t)platform_ARB_load(#n))? 1 : printf("no "#n" found.\n") & 0;

   GB_GL(PFNGLGENBUFFERSPROC,glGenBuffers);
   GB_GL(PFNGLBINDBUFFERPROC,glBindBuffer);
   GB_GL(PFNGLBUFFERDATAPROC,glBufferData);
   GB_GL(PFNGLBUFFERSUBDATAPROC,glBufferSubData);
   GB_GL(PFNGLDELETEBUFFERSPROC,glDeleteBuffers);
   GB_GL(PFNGLGETBUFFERPARAMETERIVPROC,glGetBufferParameteriv);
   GB_GL(PFNGLMAPBUFFERPROC,glMapBuffer);
   GB_GL(PFNGLUNMAPBUFFERPROC,glUnmapBuffer);
   GB_GL(PFNGLGETBUFFERSUBDATAPROC,glGetBufferSubData);
   GB_GL(PFNGLISBUFFERPROC,glIsBuffer);
   GB_GL(PFNGLDRAWRANGEELEMENTSPROC,glDrawRangeElements);
   GB_GL(PFNGLSHADERSOURCEPROC,glShaderSource);
   GB_GL(PFNGLCOMPILESHADERPROC,glCompileShader);
   GB_GL(PFNGLCREATESHADERPROC,glCreateShader);
   GB_GL(PFNGLGETSHADERIVPROC,glGetShaderiv);
   GB_GL(PFNGLGETSHADERINFOLOGPROC,glGetShaderInfoLog);
   GB_GL(PFNGLCREATEPROGRAMPROC,glCreateProgram);
   GB_GL(PFNGLATTACHSHADERPROC,glAttachShader);
   GB_GL(PFNGLLINKPROGRAMPROC,glLinkProgram);
   GB_GL(PFNGLGETPROGRAMIVPROC,glGetProgramiv);
   GB_GL(PFNGLGETPROGRAMINFOLOGPROC,glGetProgramInfoLog);
   GB_GL(PFNGLGETUNIFORMLOCATIONPROC,glGetUniformLocation);
   GB_GL(PFNGLUSEPROGRAMPROC,glUseProgram);
   GB_GL(PFNGLUNIFORM2FPROC,glUniform2f);
   GB_GL(PFNGLUNIFORM3FPROC,glUniform3f);
   GB_GL(PFNGLDELETEPROGRAMPROC,glDeleteProgram);
   GB_GL(PFNGLDELETESHADERPROC,glDeleteShader);
   GB_GL(PFNGLDETACHSHADERPROC,glDetachShader);
   GB_GL(PFNGLISPROGRAMPROC,glIsProgram);
   GB_GL(PFNGLISSHADERPROC,glIsShader);
   GB_GL(PFNGLGETATTRIBLOCATIONPROC,glGetAttribLocation);
   GB_GL(PFNGLUNIFORMMATRIX4FVPROC,glUniformMatrix4fv);
   GB_GL(PFNGLVERTEXATTRIBPOINTERPROC,glVertexAttribPointer);
   GB_GL(PFNGLVALIDATEPROGRAMPROC,glValidateProgram);
   GB_GL(PFNGLENABLEVERTEXATTRIBARRAYPROC,glEnableVertexAttribArray);
   GB_GL(PFNGLDISABLEVERTEXATTRIBARRAYPROC,glDisableVertexAttribArray);
   GB_GL(PFNGLUNIFORM4FPROC,glUniform4f);
   GB_GL(PFNGLUNIFORM1FVPROC,glUniform1fv);
   GB_GL(PFNGLUNIFORM2FVPROC,glUniform2fv);
   GB_GL(PFNGLUNIFORM3FVPROC,glUniform3fv);
   GB_GL(PFNGLUNIFORM4FVPROC,glUniform4fv);
   GB_GL(PFNGLUNIFORM1IVPROC,glUniform1iv);
   GB_GL(PFNGLUNIFORM2IVPROC,glUniform2iv);
   GB_GL(PFNGLUNIFORM3IVPROC,glUniform3iv);
   GB_GL(PFNGLUNIFORM4IVPROC,glUniform4iv);
   GB_GL(PFNGLUNIFORM1FPROC,glUniform1f);
   GB_GL(PFNGLUNIFORM1IPROC,glUniform1i);
   GB_GL(PFNGLUNIFORM2IPROC,glUniform2i);
   GB_GL(PFNGLUNIFORM3IPROC,glUniform3i);
   GB_GL(PFNGLUNIFORM4IPROC,glUniform4i);
   GB_GL(PFNGLUNIFORMMATRIX2FVPROC,glUniformMatrix2fv);
   GB_GL(PFNGLUNIFORMMATRIX3FVPROC,glUniformMatrix3fv);
   GB_GL(PFNGLBINDATTRIBLOCATIONPROC,glBindAttribLocation);
   GB_GL(PFNGLACTIVETEXTUREPROC,glActiveTexture);


   return ok;
}


#ifdef __cplusplus
}
#endif
