/* Jerry Dekker
   Purdue Visulization Lab

Simple generic stereoscopic rendering environment
**** Not fully functional -- quad screen problem unresolved.
**** Additionally, tracker support is not integrated in yet.
**** Source code cleaned up on March 26, 2009.

Model viewer with mouse/keyboard roaming.
The keyboard arrow keys control eye position.
Mouse position controls eye XZ and XY rotation angles. */

#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"
#include <errno.h>
#define USING_GLFW_RUNTIME_LD
#include "ext_libs.h"
#include "display.h"
//#define DEBUG_ALL

#define FPS_UPDATE            1.0  /* seconds */

/* Trig definitions */
#define HALF_PI               1.5707963f
#define PI                    3.1415927f
#define THREE_HALVES_PI       4.7123890f
#define TWO_PI                6.2831853f

/* Screen modes */
enum { WINDOWED_MODE, FULLSCREEN_MODE };

/* View modes */
enum {
   NORMAL_VIEWING,
   ANAGLYPH,
   HORIZ_INTERLACED,
   VERT_INTERLACED,
   CHECKERBOARD,
   STEREO_PAIR,
   HARDWARE_STEREO,
   HARDWARE_QUAD,
   STEREO_QUAD
};

static char mode_names[9][30] = {
   "None",
   "Anaglyph",
   "      Row\nInterlaced",
   "  Column\nInterlaced",
   "Checker",
   "Stereo Pair",
   "Hardware\n   Stereo",
   "Hardware Quad",
   "Stereo Quad"
};

/* Viewport designators */
enum {
   VIEW_FRONT_CENTER,
   VIEW_FRONT_RIGHT,
   VIEW_FRONT_LEFT,
   VIEW_BACK_RIGHT,
   VIEW_BACK_LEFT
};

/* Rendering system functions. */
static void init(int);
static int  open_display(int);
static void GLFWCALL reshape(int ,int);
static void set_stencil(int ,int);
static int  GLFWCALL close_window(void);
static void GLFWCALL keyboard (int, int);
static void GLFWCALL active_mouse(int, int);
static void GLFWCALL mouse_wheel(int);
static void toggle_back_buffers(void);
static void toggle_front_buffers(void);
static void pre_render(void);
static void update_view(void);
static void render(void);

/* Window environment */
static int GLFW_run              = 0;
static int window_width          = 1000;
static int window_height         = 1000;
static int fullscreen_width      = 1920;
static int fullscreen_height     = 1200;
static int window_mode           = FULLSCREEN_MODE;
static int last_height           = 1000;
static int last_width            = 1000;

/* Buffer targets */
static unsigned int left_front   = GL_BACK_LEFT;
static unsigned int right_front  = GL_BACK_RIGHT;
static unsigned int left_bottom  = GL_BACK_RIGHT;
static unsigned int right_bottom = GL_BACK_LEFT;

/* Viewport info */
static int view_width            = 0;
static int view_height           = 0;
static int view_mode             = NORMAL_VIEWING;

/* View frustum info */
static float flat_clip           = 0;
static float clip_left           = 0;
static float clip_right          = 0;
static float clip_span           = 0;
static float near_clip           = 0.05f;
static float far_clip            = 20000.0f;
static float aspect_ratio        = 0;
static float half_fov            = 0.8726f / 2;
static float GLFW_eye_seperation = -0.001800;
static float lin_eye_sep         = -0.009;
static float GLFW_focal_length   = 30.0;
static float lowview_rotate      = 0;
static double left_frustum[16];
static double center_frustum[16];
static double right_frustum[16];
static double identity_matrix[16] = {
   1,0,0,0,
   0,1,0,0,
   0,0,1,0,
   0,0,0,1
};
static int max_viewport;
static int min_viewport;
static unsigned int GLFW_mode_limit = STEREO_PAIR;
/* static GLFWvidmode v_mode[50]; */

/* Control configuration */
static int set_mouse             = 1;
static int last_x                = 0;
static int last_y                = 0;
static float move_scale          = 0.0001f;
static float look_scale          = 0.001f;

/* New additons according to display.h... */
display_properties frame_state;

/* Application function pointers. */
static void (*app_preframe)(void) = 0;
static void (*app_context)(void)  = 0;
static void (*app_draw)(void)     = 0;

static unsigned int frame_count   = 0;
static double       fps_dt_accum  = 0;


void (*display_set_stereo_mode)(unsigned int mode_number);
void (*display_set_clip_plane)(float gl_z_value);
void (*display_set_eye_seperation)(float value);
void (*display_set_focal_plane)(float gl_z_value);

void set_stereo_mode(unsigned int mode_number);
void set_clip_plane(float gl_z_value);
void set_eye_seperation(float value);
void set_focal_plane(float gl_z_value);

void display_prepare(int argc, char **argv)
{
   unsigned n = 1;
   GLFWvidmode desktop;

   if (errno > 0) {
      perror("Startup errno cleanup; setting to zero, was");
      errno = 0;
   }

   #ifdef USING_GLFW_RUNTIME_LD
      if (!ext_load_glfw()) {
         printf("GLFW dynamic library was not found. Exiting...\n");
         exit(1);
      }
   #endif

   while (n < argc) {
      if (argv[n] == strstr(argv[n],"hardware")) {
         GLFW_mode_limit = HARDWARE_STEREO;
         break;
      }
      n++;
   }

   glfwInit();
   glfwGetDesktopMode(&desktop);
   fullscreen_width  = desktop.Width;
   fullscreen_height = desktop.Height;
   printf("Using %d x %d\n",fullscreen_width,fullscreen_height);

   if (GLFW_mode_limit == HARDWARE_STEREO)
      glfwOpenWindowHint(GLFW_STEREO,1);

   window_width  = fullscreen_width;
   window_height = fullscreen_height;

   if (open_display(GLFW_FULLSCREEN) == GL_FALSE) {
      printf("No Stereo Hardware Detected.\n");
      glfwOpenWindowHint(GLFW_STEREO,0);
      open_display(GLFW_FULLSCREEN);
      GLFW_mode_limit = STEREO_PAIR;
   }

   glfwSetWindowCloseCallback(&close_window);
   glfwSetWindowSizeCallback(&reshape);
   glfwSetKeyCallback(&keyboard);
   glfwEnable(GLFW_KEY_REPEAT);
   glfwSetMouseButtonCallback(&active_mouse);
   glfwSetMouseWheelCallback(&mouse_wheel);
   glfwGetMousePos(&last_x,&last_y);

   display_set_stereo_mode    = &set_stereo_mode;
   display_set_clip_plane     = &set_clip_plane;
   display_set_eye_seperation = &set_eye_seperation;
   display_set_focal_plane    = &set_focal_plane;

   frame_state.stereo_modes    = GLFW_mode_limit;
}

void display_set_context_renew_function(void_callback *ptr)
{
   app_context = ptr;
}

void display_set_preframe_function(void_callback *ptr)
{
   app_preframe = ptr;
}

void display_set_display_function(void_callback *ptr)
{
   app_draw = ptr;
}

void set_stereo_mode(unsigned int mode_number)
{
   if (mode_number > GLFW_mode_limit)
      return;

   view_mode = mode_number;
   if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
      || view_mode == CHECKERBOARD)
      set_stencil(window_width,window_height);
   update_view();
}

void set_clip_plane(float gl_z_value)
{
   if (gl_z_value <= 0)
      gl_z_value = .000001;

   near_clip = gl_z_value;
   update_view();
}

void set_eye_seperation(float value)
{
	GLFW_eye_seperation = value;
   update_view();
}

void set_focal_plane(float gl_z_value)
{
   GLFW_focal_length = gl_z_value;
   update_view();
}


void display_run_loop(void)
{
   GLFW_run = 1;

   if (app_context)
      app_context();

   /* Main rendering loop. */
   while (GLFW_run) {
      pre_render();
      render();
      glfwSwapBuffers();
   }

   /* Ending...clean up. */
   glfwTerminate();
}

void display_shutdown(void)
{
   GLFW_run = 0;
}

void init(int type)
{
   glfwTerminate();
   glfwInit();

   if (GLFW_mode_limit == HARDWARE_STEREO)
      glfwOpenWindowHint(GLFW_STEREO,1);

   switch (type) {
      case WINDOWED_MODE:
         window_width  = last_width;
         window_height = last_height;

         if (open_display(GLFW_WINDOW) == GL_FALSE) {
            printf("Window mode not supported...\n");
            GLFW_run = 0;
         }
         glfwSetWindowPos(0,0);
         glfwSetWindowTitle("The Planck Mission Simulation");
         glfwDisable(GLFW_MOUSE_CURSOR);
         break;
      case FULLSCREEN_MODE:
         last_width    = window_width;
         last_height   = window_height;
         window_width  = fullscreen_width;
         window_height = fullscreen_height;

         if (open_display(GLFW_FULLSCREEN) == GL_FALSE) {
            printf("Fullscreen mode not supported...\n");
            GLFW_run = 0;
         }
         break;
   }

   glfwSetWindowCloseCallback(&close_window);
   glfwSetWindowSizeCallback(&reshape);
   glfwSetKeyCallback(&keyboard);
   glfwEnable(GLFW_KEY_REPEAT);
   glfwSetMouseButtonCallback(&active_mouse);
   glfwSetMouseWheelCallback(&mouse_wheel);
   glfwGetMousePos(&last_x,&last_y);

   if (app_context)
      app_context();
}

int open_display(int mode)
{
   return glfwOpenWindow(window_width,window_height,8,8,8,8,32,1,mode);
}

int GLFWCALL close_window(void)
{
   GLFW_run = 0;
   return 1;
}

void GLFWCALL reshape(int width,int height)
{
   /* Resize display related objects. */
   window_width  = width;
   window_height = height;
   if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
      || view_mode == CHECKERBOARD)
      set_stencil(window_width,window_height);
   update_view();
}

void set_stencil(int width,int height)
{
	#define MAX_BM 2048

   /* Assuming stencil buffer of depth 1... */
   unsigned char *bm;
   int scanline_count, max_dim = (width > height) ? width : height;
   const char* ATI = strstr(glGetString(GL_VENDOR),"ATI");

   /* Set up the raster position. */
   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
   glOrtho(0,width,0,height,0,1);
   glViewport(0,0,width,height);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();
   glRasterPos2i(0,0);

   /* This should be the ONLY place the stencil buffer is cleared, as we want
   to reuse the raster pattern for each frame. */
   glStencilMask(~0);
   glClearStencil(0);
   glClear(GL_STENCIL_BUFFER_BIT);
   glStencilFunc(GL_ALWAYS,1,1);
   glEnable(GL_STENCIL_TEST);

   if (ATI) {

	   bm = (unsigned char*)malloc(max_dim * sizeof(int));
  	 	if (!bm) {
	  	 	glDisable(GL_STENCIL_TEST);
   		glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
      	return;
   	}

   	memset(bm,0xFF,max_dim * sizeof(int));

	   if (view_mode == HORIZ_INTERLACED || view_mode == CHECKERBOARD) {
		   glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
		   scanline_count = 0;
      	while (scanline_count < height) {
         	if (scanline_count%2) {
         		glRasterPos2i(0,scanline_count);
         		glDrawPixels(width,1,GL_RGBA,GL_INT,bm);
      		}
         	scanline_count++;
      	}
   	}

   	if (view_mode == VERT_INTERLACED || view_mode == CHECKERBOARD) {
	      glStencilOp(GL_INVERT,GL_INVERT,GL_INVERT);
	      scanline_count = 0;
      	while (scanline_count < width) {
         	if (scanline_count%2) {
         		glRasterPos2i(scanline_count,0);
         		glDrawPixels(1,height,GL_RGBA,GL_INT,bm);
      		}
         	scanline_count++;
      	}
   	}
   }
   else {

	   bm = (unsigned char*)malloc(MAX_BM/8 * MAX_BM * sizeof(unsigned char));
   	if (!bm) {
      	glDisable(GL_STENCIL_TEST);
   		glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
      	return;
   	}

   	memset(bm,0xFF,MAX_BM/8 * MAX_BM);

	   if (view_mode == HORIZ_INTERLACED || view_mode == CHECKERBOARD) {
		   scanline_count = 0;
	      while (scanline_count < height) {
	         if (scanline_count%2)
	            memset(bm + scanline_count * MAX_BM/8,0xFF,MAX_BM/8);
	         else
	            memset(bm + scanline_count * MAX_BM/8,0,MAX_BM/8);
	         scanline_count++;
	      }
	      glStencilOp(GL_REPLACE,GL_REPLACE,GL_REPLACE);
	      glBitmap(MAX_BM,MAX_BM,0,0,0,0,bm);
	   }

	   if (view_mode == VERT_INTERLACED || view_mode == CHECKERBOARD) {
	      memset(bm,0xAA,MAX_BM/8 * MAX_BM);
	      glStencilOp(GL_INVERT,GL_INVERT,GL_INVERT);
	      glBitmap(MAX_BM,MAX_BM,0,0,0,0,bm);
	   }
   }

   free(bm);
   glDisable(GL_STENCIL_TEST);
   glStencilOp(GL_KEEP,GL_KEEP,GL_KEEP);
}

void GLFWCALL keyboard (int key, int state)
{
   switch (state) {
      case GLFW_PRESS:
         frame_state.key = key;
         switch (key) {
            case GLFW_KEY_ESC:
               display_shutdown();
               break;
            case GLFW_KEY_HOME:
               break;
            case GLFW_KEY_PAGEUP:
               /* GLFW_eye_seperation += .00001f;
               update_view(); */
               break;
            case GLFW_KEY_PAGEDOWN:
               /* GLFW_eye_seperation -= .00001f;
               update_view(); */
               break;
            case GLFW_KEY_F1:
               break;
            case GLFW_KEY_F2:
               if (window_mode == FULLSCREEN_MODE)
                  window_mode = WINDOWED_MODE;
               else window_mode = FULLSCREEN_MODE;
               init(window_mode);
               break;
            case GLFW_KEY_F3:
               break;
            case GLFW_KEY_F4:
               break;
            case GLFW_KEY_F5:
               break;
            case GLFW_KEY_F6:
               break;
            case GLFW_KEY_F7:
               break;
            case GLFW_KEY_F8:
               break;
            case GLFW_KEY_F9:
               toggle_back_buffers();
               break;
            case GLFW_KEY_F10:
              /* GLFW_eye_seperation = -GLFW_eye_seperation;
               update_view(); */
               break;
            case GLFW_KEY_F11:
              /* GLFW_focal_length -= .1;
               if (GLFW_focal_length < .1)
                  GLFW_focal_length = .1;
               update_view(); */
               break;
            case GLFW_KEY_F12:
              /* GLFW_focal_length += .1;
               update_view(); */
               break;
            case GLFW_KEY_DEL:
               if (set_mouse = !set_mouse)
                  glfwDisable(GLFW_MOUSE_CURSOR);
               else glfwEnable(GLFW_MOUSE_CURSOR);
               break;
            default:
            break;
         }
      	break;
      case GLFW_RELEASE:
      	frame_state.key = 0;
         break;
   }
}

void GLFWCALL active_mouse(int button, int state)
{
   switch (button) {
      case GLFW_MOUSE_BUTTON_LEFT:
         switch (state) {
            case GLFW_PRESS:
               break;
            case GLFW_RELEASE:
               break;
         }
         break;
      case GLFW_MOUSE_BUTTON_RIGHT:
         switch (state) {
            case GLFW_PRESS:
               break;
            case GLFW_RELEASE:
               break;
         }
         break;
      case GLFW_MOUSE_BUTTON_MIDDLE:
         switch (state) {
            case GLFW_PRESS:
               break;
            case GLFW_RELEASE:
               break;
         }
         break;
   }
}

void GLFWCALL mouse_wheel(int position)
{
   return;
}

/* Functions which reverse left and right buffers for hardware stereo
rendering. This has the effect of reversing eye pairs. */

void toggle_back_buffers()
{
   unsigned int temp = left_bottom;

   left_bottom       = right_bottom;
   right_bottom      = temp;
}

void toggle_front_buffers()
{
   unsigned int temp = left_front;

   left_front        = right_front;
   right_front       = temp;
}

void pre_render(void)
{
   static int x, y, dx, dy;
   float up_down = 0, left_right = 0, move = 0;
   static float dt, *rot, *trans, rev = 0, sin_r0, cos_r0, sin_r1, cos_r1;
   unsigned int input = 0;

   /* Time, metrics, and synchronization... */
   dt = frame_state.dt = platform_delta_timer();

   fps_dt_accum += frame_state.dt;
   frame_count++;
   if (fps_dt_accum >= FPS_UPDATE) {
      frame_state.fps = frame_count/fps_dt_accum;
      frame_count     = 0;
      fps_dt_accum    = 0;
   }

   /* Input control update... */
   glfwGetMousePos(&x,&y);
   dx     = last_x - x;
   dy     = last_y - y;
   last_x = x;
   last_y = y;
   rot    = frame_state.rotation;
   trans  = frame_state.position;

   if (glfwGetKey(GLFW_KEY_UP))
      up_down = move = move_scale;
   if (glfwGetKey(GLFW_KEY_DOWN))
      up_down = move = -move_scale;
   if (glfwGetKey(GLFW_KEY_RIGHT))
      left_right = move = move_scale;
   if (glfwGetKey(GLFW_KEY_LEFT))
      left_right = move = -move_scale;

  /* if (glfwGetKey('1'))
      ctl_button_set(1,input);
   if (glfwGetKey('3'))
      ctl_button_set(3,input);
   if (glfwGetKey('4'))
      ctl_button_set(4,input);
   if (glfwGetKey('5'))
      ctl_button_set(5,input);
   if (glfwGetKey(GLFW_KEY_F1))
      ctl_button_set(10,input);
   if (glfwGetKey(GLFW_KEY_F3))
      ctl_button_set(11,input); */

   /* Simulate VRJ behavior. */
   if (move)
      ctl_button_set(0,input);
   if (dx || dy)
      ctl_button_set(2,input);

   if (glfwGetMouseButton(GLFW_MOUSE_BUTTON_LEFT))
      ctl_button_set(5,input);
   if (glfwGetMouseButton(GLFW_MOUSE_BUTTON_RIGHT))
      ctl_button_set(4,input);

   frame_state.pointer[0] = -dx;
   frame_state.pointer[1] = dy;
   frame_state.button     = input;

   if (rot && trans) {

      if (dx) {
         rev = (rot[1] - dx * look_scale)/TWO_PI;
         rot[1] = TWO_PI * ((rev >= 0) ? rev - (int)rev : 1 + rev - (int)rev);
      }

      if (dy) {
         rev = (rot[0] - dy * look_scale)/TWO_PI;
         rot[0] = TWO_PI * ((rev >= 0) ? rev - (int)rev : 1 + rev - (int)rev);

         if (rot[0] > HALF_PI && rot[0] <= PI)
            rot[0] = HALF_PI;
         if (rot[0] > PI && rot[0] < THREE_HALVES_PI)
            rot[0] = THREE_HALVES_PI;
      }

      /* Calculate the marginal movement hypotenuse. */
      if (move) {
         sin_r0 = sin(rot[0]);
         cos_r0 = cos(rot[0]);
         sin_r1 = sin(rot[1]);
         cos_r1 = cos(rot[1]);

         trans[0] -= (up_down * sin_r1 + left_right * cos_r1) * cos_r0;
         trans[1] += up_down * sin_r0;
         trans[2] += (up_down * cos_r1 - left_right * sin_r1) * cos_r0;
      }

      move_scale = (move) ? (move_scale += 0.1 * dt) : 0.0001;
   }

   if (app_preframe)
      app_preframe();

   #ifdef DEBUG_ALL
   {
      unsigned error_check = glGetError();

      switch (error_check) {
         case GL_NO_ERROR:
            break;
         case GL_INVALID_ENUM:
            printf("Last GL_ERROR: invalid enumerator.\n");
            break;
         case GL_INVALID_VALUE:
            printf("Last GL_ERROR: invalid value.\n");
            break;
         case GL_INVALID_OPERATION:
            printf("Last GL_ERROR: invalid operation.\n");
            break;
         case GL_STACK_OVERFLOW:
            printf("Last GL_ERROR: stack overflow.\n");
            break;
         case GL_STACK_UNDERFLOW:
            printf("Last GL_ERROR: stack underflow.\n");
            break;
         case GL_OUT_OF_MEMORY:
            printf("Last GL_ERROR: out of memory.\n");
            break;
         default:
            printf("Last GL_ERROR: unknown.\n");
      }

      if (errno > 0) {
         perror("C runtime error detected");
         errno = 0;
      }

      printf(
         "eye seperation: %f  focal length: %f            \r",
         GLFW_eye_seperation,
         GLFW_focal_length);
   }
   #endif
}

/* This is called whenever any variable which affects the window / viewport
configuration changes. This can be a window resize, a stereo mode change,
an eye seperation modification, etc. */

void update_view(void)
{
   /* The "eye_toe" variable sets the theoretical focal point
   (even though each eye view is really parallel) by which the view
   frustums are warped. This effectively creates the illusion of
   specific object focus.

   The "lowview_rotate" rotates the low view (floor) cameras up
   so that the edge of the frustum meets the bottom of the top views.
   This is not complete--additional lower frustum warping is necessary.

   This partitions the window into view port areas based on the viewing
   configuration. */

   float frust_shift;
   int viewport[4];
   double *use_frust, lower_left[3] = {0,0,0}, upper_right[3] = {0,0,0};

   glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
   glDisable(GL_STENCIL_TEST);

   switch (view_mode) {
      case HORIZ_INTERLACED:
      case VERT_INTERLACED:
      case CHECKERBOARD:
         glEnable(GL_STENCIL_TEST);
      case NORMAL_VIEWING:
      case HARDWARE_STEREO:
      case ANAGLYPH:
         view_width = window_width;
         view_height = window_height;
         break;
      case STEREO_QUAD:
         view_width = window_width/2;
         view_height = window_height/2;
         break;
      case STEREO_PAIR:
      case HARDWARE_QUAD:
         view_width = window_width/2;
         view_height = window_height;
         break;
   }

   lin_eye_sep    = -0.5 * GLFW_eye_seperation;
   aspect_ratio   = (float)view_height / view_width;
   flat_clip      = tan(half_fov) * near_clip;
   frust_shift    = lin_eye_sep * near_clip / GLFW_focal_length;
   clip_left      = -flat_clip + frust_shift;
   clip_right     = flat_clip + frust_shift;
   clip_span      = flat_clip * aspect_ratio;
   lowview_rotate = 2 * atan(clip_span/near_clip) * 57.2958;

   /* Create and store the current frustum and  movement matrices... */

   glMatrixMode(GL_PROJECTION);

   switch (view_mode) {
      case NORMAL_VIEWING:
         glLoadIdentity();
         glFrustum(-flat_clip,flat_clip,-clip_span,clip_span,near_clip,far_clip);
         glGetDoublev(GL_PROJECTION_MATRIX,center_frustum);
         break;
      case ANAGLYPH:
      case HARDWARE_STEREO:
      case STEREO_PAIR:
      case HARDWARE_QUAD:
      case STEREO_QUAD:
      case HORIZ_INTERLACED:
      case VERT_INTERLACED:
      case CHECKERBOARD:
         glLoadIdentity();
         glFrustum(clip_left,clip_right,-clip_span,clip_span,near_clip,far_clip);
         glGetDoublev(GL_PROJECTION_MATRIX,right_frustum);
         glLoadIdentity();
         glFrustum(-clip_right,-clip_left,-clip_span,clip_span,near_clip,far_clip);
         glGetDoublev(GL_PROJECTION_MATRIX,left_frustum);
         break;
   }

   switch (view_mode) {
      case NORMAL_VIEWING:
         min_viewport = VIEW_FRONT_CENTER;
         max_viewport = VIEW_FRONT_CENTER;
         break;
      case ANAGLYPH:
      case HORIZ_INTERLACED:
      case VERT_INTERLACED:
      case CHECKERBOARD:
         min_viewport = VIEW_FRONT_RIGHT;
         max_viewport =  VIEW_FRONT_LEFT;
         break;
      case HARDWARE_STEREO:
      case STEREO_PAIR:
         min_viewport = VIEW_FRONT_RIGHT;
         max_viewport = VIEW_FRONT_LEFT;
         break;
      case HARDWARE_QUAD:
      case STEREO_QUAD:
         min_viewport = VIEW_FRONT_RIGHT;
         max_viewport =  VIEW_BACK_LEFT;
         break;
   }

   /* use_frust = (view_mode) ? right_frustum : center_frustum; */
   use_frust = center_frustum;

   /* Calculate the screen overlay configuration... */
   viewport[0] = viewport[1] = 0;
   viewport[2] = view_width;
   viewport[3] = view_height;

   gluUnProject(
      0,0,.998999,                                    /* Coordinate: lower left.*/
      identity_matrix,use_frust,viewport,             /* GL configuration. */
      lower_left,lower_left + 1,lower_left + 2);      /* Values out. */

   gluUnProject(
      (double)view_width,(double)view_height,.998999, /* Coordinate: upper right */
      identity_matrix,use_frust,viewport,             /* GL configuration. */
      upper_right,upper_right + 1,upper_right + 2);   /* Values out. */

   /* Export  external properties... */
   frame_state.window_pixels[0]  = window_width;
   frame_state.window_pixels[1]  = window_height;
   frame_state.overlay_area[0]   = lower_left[0];
   frame_state.overlay_area[1]   = lower_left[1];
   frame_state.overlay_area[2]   = upper_right[0] - lower_left[0];
   frame_state.overlay_area[3]   = upper_right[1] - lower_left[1];
   frame_state.overlay_area[4]   = upper_right[2];
   frame_state.stereo_modes      = GLFW_mode_limit;
   frame_state.current_mode      = view_mode;
   frame_state.stereo_mode_names = mode_names;
   frame_state.eye_seperation    = GLFW_eye_seperation;
   frame_state.focal_plane       = GLFW_focal_length;
   frame_state.clipping_plane    = near_clip;
}

void render(void)
{
   int viewport = min_viewport;

   /* Begin main drawing section. */
   if (view_mode == ANAGLYPH)
      glColorMask(GL_TRUE,GL_TRUE,GL_TRUE,GL_TRUE);
   glDrawBuffer(GL_BACK);
   glClear(GL_COLOR_BUFFER_BIT);

   while (viewport <= max_viewport) {

      glClear(GL_DEPTH_BUFFER_BIT);

      /* ViewPort and Buffer setup */
      switch (viewport) {
         case VIEW_FRONT_CENTER:
         case VIEW_FRONT_RIGHT:
            switch (view_mode) {
               case HARDWARE_QUAD:
               case HARDWARE_STEREO:
                  glDrawBuffer(left_front);
                  break;
            }
            glViewport(0,0,view_width,view_height);
            break;
         case VIEW_BACK_RIGHT:
            switch (view_mode) {
               case HARDWARE_QUAD:
                  glDrawBuffer(left_bottom);
                  glViewport(view_width,0,view_width,view_height);
                  break;
               case STEREO_QUAD:
                  glViewport(0,view_height,view_width,view_height);
                  break;
            }
            break;
         case VIEW_FRONT_LEFT:
            switch (view_mode) {
               case STEREO_PAIR:
               case STEREO_QUAD:
                  glViewport(view_width,0,view_width,view_height);
                  break;
               case HARDWARE_STEREO:
               case HARDWARE_QUAD:
                  glDrawBuffer(right_front);
                  glViewport(0,0,view_width,view_height);
                  break;
            }
            break;
         case VIEW_BACK_LEFT:
            switch (view_mode) {
               case HARDWARE_QUAD:
                  glDrawBuffer(right_bottom);
                  glViewport(view_width,0,view_width,view_height);
                  break;
               case STEREO_QUAD:
                  glViewport(view_width,view_height,view_width,view_height);
                  break;
            }
            break;
      } /* End viewport and buffer setup. */

      /* Frustum setup */
      glMatrixMode(GL_PROJECTION);
      switch (viewport) {
         case VIEW_FRONT_CENTER:
            glLoadMatrixd(center_frustum);
            break;
         case VIEW_FRONT_RIGHT:
            glLoadMatrixd(right_frustum);
            break;
         case VIEW_BACK_RIGHT:
            glLoadMatrixd(right_frustum);
            glRotatef(lowview_rotate,1,0,0);
            break;
         case VIEW_FRONT_LEFT:
            glLoadMatrixd(left_frustum);
            break;
         case VIEW_BACK_LEFT:
            glLoadMatrixd(left_frustum);
            glRotatef(lowview_rotate,1,0,0);
            break;
      }/* End frustum setup. */

      /* Eye offset and position setup */
      switch (viewport) {
         case VIEW_FRONT_CENTER:
            break;
         case VIEW_FRONT_RIGHT:
         case VIEW_BACK_RIGHT:
            if (view_mode == ANAGLYPH)
               glColorMask(GL_FALSE,GL_TRUE,GL_TRUE,GL_FALSE);
            if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
               || view_mode == CHECKERBOARD)
               glStencilFunc(GL_EQUAL,1,1);
            glTranslatef(lin_eye_sep,0,0);
            break;
         case VIEW_FRONT_LEFT:
         case VIEW_BACK_LEFT:
            if (view_mode == ANAGLYPH)
               glColorMask(GL_TRUE,GL_FALSE,GL_FALSE,GL_FALSE);
            if (view_mode == HORIZ_INTERLACED || view_mode == VERT_INTERLACED
               || view_mode == CHECKERBOARD)
               glStencilFunc(GL_NOTEQUAL,1,1);
            glTranslatef(-lin_eye_sep,0,0);
            break;
      } /*  End of eye offset and position setup */

      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();

      if (app_draw)
         app_draw();

      viewport++;
   }
}

#ifdef __cplusplus
}
#endif
