#ifndef __GLTF_H__
#define __GLTF_H__

/* Open GL Texture Mapped Font Library (GLTF) */

#ifdef __cplusplus
extern "C" {
#endif

#include "renderer_api.h"
#include "gluf.h"


#define DEFAULT_FONT_DIR "./data/fonts"

typedef struct gl_tex_font gl_tex_font;

struct gl_tex_font {

   void (*erase)(gl_tex_font *tf);
   void (*draw)(gl_tex_font *tf, const char *txt, mat44 *projection, mat44 *modelview);

   /* All arrays are directly indexed by ASCII character encoding... */

   short x[256];           /* Texture u coordinate in pixels.*/
   short y[256];           /* Texture v coordinate in pixels.*/
   short w[256];           /* Texture glyph width in pixels.*/
   short h[256];           /* Texture glyph height in pixels.*/
   short pad[256];         /* Post padding for glyph in pixels.*/
   short x_off[256];       /* Required texture fragment x offset in pixels.*/
   short y_off[256];       /* Required texture fragment y offset in pixels.*/
   short kern[256][256];   /* Kerning table.*/
   short line_h;           /* Line height in pixels. */
   short max_h;            /* Height of largest character. */
   short tex_w;            /* Reported texture width in pixels. */
   short tex_h;            /* Reported texture height in pixels. */

   char *fnt_tex_file;     /* Font texture file name. */
   gluf_bitmap *fnt_image; /* Texture image data. */
   unsigned tex_id;        /* Texture id. */

   vec2 *vtx;          /* Vertex array. */
   vec2 *tex;          /* Texture coordinate array */
   unsigned vtx_sz;        /* Current size of arrays. */
   unsigned use_vtx;       /* Actual count of array which pertains to most
                              recent operation. */
   unsigned tab_stop;      /* Drawn document tab size. */
   unsigned cpl;           /* Characters per line; cpl = 0 is auto formating. */
   unsigned justify_top;   /* Justify text to top of box; default is bottom */
   unsigned render;        /* Render, or just generate arrays: default true. */
   unsigned centered;      /* Center the text based on calculated line width. */
   unsigned pseudo_mono;   /* Attempt a monospace-like line usage pattern... */
   unsigned use_w;         /* Assumed constant width for pseudo_mono. */

   float    color[4];      /* Color to use if drawn. */

};

gl_tex_font *create_texture_font(const char *font_file);


#ifdef __cplusplus
}
#endif

#endif /* __GLTF_H__ */