#ifdef __cplusplus
extern "C" {
#endif

#define USING_OPENGL
#include "platform.h"
#define USING_FREEIMAGE_RUNTIME_LD
#include "ext_libs.h"
#include "gluf.h"
#include "lz4.h"
#include "renderer_api.h"

#define TWO_PI       6.2831853
#define DIVISIONS    360


typedef struct {
   char        *image_file;
   gluf_bitmap **loc_ptr;
}
async_param;

static t_func_t get_bm_async(void *param);

static char *buffer = 0;

static unsigned test = 0;
void gluf_image_mix_in(char *dest, char *src, int x, int y, unsigned order)
{
   int n = 0, m = 0, total = x * y;
   gluf_pixel *s = (gluf_pixel*)src, *d = (gluf_pixel*)dest;
   gluf_pixel *s1, *d1;
   unsigned flip = order;

   if (!d || !s || !total)
      return;

//    while (n < total) {
//       s1 = s + n;
//       d1 = d + n++;
//
//       d1->r = (d1->r >> 1) + (s1->r >> 1);
//       d1->g = (d1->g >> 1) + (s1->g >> 1);
//       d1->b = (d1->b >> 1) + (s1->b >> 1);
//
//       //*(d + n++) = d1;
//    }

   while (m < y) {
      n = flip;
      s1 = s + m * x;
      d1 = d + m * x;
      while (n < x) {
         d1[n] = s1[n];
         //d[n].i = test;
         n++;
         n++;
      }
      flip = !flip;
      m++;
   }
}

void gluf_free_rle(gluf_rle_data *rle)
{
   if (!rle)
      return;

   if (rle->data)
      free(rle->data);

   free(rle);
}

void gluf_image_encode_fast_buffer_size(unsigned sz)
{
   if (buffer)
      free(buffer);

   if (sz)
      buffer = (char*)malloc(sz);

   printf("resize: %u\n",sz);
}

gluf_rle_data *gluf_image_encode_fast(char *data, int x, int y)
{
   int n = 0, m = 0, out_sz, end, sz, lz4_sz;
   gluf_pixel *i, *o;
   char *out;
   gluf_rle_data *rle;

   if (!data || !buffer || !x || !y) {
      printf("Problems...\n");
      return 0;
   }

   end = x * y;
   sz  = end * sizeof(gluf_pixel);
//    i   = (gluf_pixel*)data;
//    o   = (gluf_pixel*)buffer;
//
//    while (n < end) {
//       o[m]   = i[n++];
//       o[m].a = 0;
//       while (i[n].r == o[m].r && i[n].g == o[m].g && i[n].b == o[m].b &&
//              n < end && o[m].a < 255) {
//          o[m].a++;
//          n++;
//       }
//       m++;
//    }
//    sz = m * sizeof(gluf_pixel);

   out = (char*)malloc(LZ4_COMPRESSBOUND(sz));
   if (out) {
      out_sz = LZ4_compress(data,out,sz);
      if (out_sz > 0) {
         lz4_sz = sz;
         sz     = out_sz;

         rle = (gluf_rle_data*)malloc(sizeof(gluf_rle_data));
         if (rle) {
            rle->data   = out;
            rle->sz     = sz;
            rle->lz4_sz = lz4_sz;
            rle->x      = x;
            rle->y      = y;

            return rle;
         }
         printf("Encoding struct allocation problem...\n");
         free(out);
         return 0;

      }
      printf("LZ4 compression error...\n");
      free(out);
      return 0;
   }
   printf("LZ4 buffer allocation problem...\n");
   return 0;
}

void gluf_image_encode(gluf_rle_data *rd)
{
   int n = 0, m = 0, out_sz, end;
   gluf_pixel *i, *o;
   char *out;

   if (!rd || !rd->x || !rd->y || !rd->data)
      return;
   end     = rd->x * rd->y;
   rd->sz  = end * sizeof(gluf_pixel);
//    i = o = (gluf_pixel*)rd->data;
//
//    while (n < end) {
//       o[m]   = i[n++];
//       o[m].a = 0;
//       while (i[n].r == o[m].r && i[n].g == o[m].g && i[n].b == o[m].b &&
//              n < end && o[m].a < 255) {
//          o[m].a++;
//          n++;
//       }
//       m++;
//    }
//
//    rd->sz = m * sizeof(gluf_pixel);

   out = (char*)malloc(LZ4_COMPRESSBOUND(rd->sz));
   if (out) {
      out_sz = LZ4_compress(rd->data,out,rd->sz);
      if (out_sz > 0) {
         rd->lz4_sz = rd->sz;
         rd->sz = out_sz;
         free(rd->data);
         rd->data = out;
         return;
      }
      printf("LZ4 compression error...\n");
      free(out);
   }
}



void gluf_image_decode(gluf_rle_data *rd)
{
   int n = 0, m = 0, a, run, end, count, read_sz;
   gluf_pixel *i, *o;
   char *out;


   if (!rd || !rd->sz || !rd->data || !rd->x || !rd->y)
      return;

   out = (char*)malloc(rd->lz4_sz);
   if (out) {

      read_sz = LZ4_uncompress(rd->data,out,rd->lz4_sz);

      if (read_sz > 0) {
         rd->sz = rd->lz4_sz;
         free(rd->data);
         rd->data = out;
      }
      else {
         printf("LZ4 decompress error...\n");
         free(out);
      }
   }

//    count = rd->x * rd->y;
//    end   = rd->sz/4;
//    o     = (gluf_pixel*)malloc(count * sizeof(gluf_pixel));
//    i     = (gluf_pixel*)rd->data;
//
//    while (n < end && m < count) {
//       a      = 0;
//       run    = i[n].a;
//       i[n].a = 1;
//
//       while (a <= run) {
//          o[m++] = i[n];
//          a++;
//       }
//       n++;
//    }
//
//    if (n != end) {
//       printf("Pixels were lost...\n");
//    }
//
//    rd->sz = count * sizeof(gluf_pixel);
//    free(rd->data);
//    rd->data = (char*)o;

}

void gluf_image_decode_invert(gluf_rle_data *rd)
{
   int n = 0, m, a, end, run, row_end, count, read_sz;
   gluf_pixel *i, *o;
   char *out;

   if (!rd || !rd->sz || !rd->data || !rd->x || !rd->y)
      return;

   out = (char*)malloc(rd->lz4_sz);
   if (out) {
      read_sz = LZ4_uncompress(rd->data,out,rd->lz4_sz);
      if (read_sz > 0) {
         rd->sz = rd->lz4_sz;
         free(rd->data);
         rd->data = out;
      }
      else {
         printf("LZ4 decompress error...\n");
         free(out);
      }
   }

//    count    = rd->x * rd->y;
//    end      = rd->sz/4;
//    o        = (gluf_pixel*)malloc(count * sizeof(gluf_pixel));
//    i        = (gluf_pixel*)rd->data;
//    row_end  = count;
//    m        = row_end - rd->x;
//
//    while (n < end) {
//       a      = 0;
//       run    = i[n].a;
//       i[n].a = 1;
//
//       while (a <= run && m >= 0) {
//          o[m++] = i[n];
//          if (m == row_end) {
//             m -= 2 * rd->x;
//             row_end = m + rd->x;
//          }
//          a++;
//       }
//       n++;
//    }
//
//    if (n != end) {
//       printf("Pixels were lost...\n");
//    }
//
//    rd->sz = count * sizeof(gluf_pixel);
//    free(rd->data);
//    rd->data = (char*)o;
}

/************************* 2d vertex array functions **************************/

void gluf_arc_2d(vec2 *v_arr, unsigned v_cnt, float begin_deg, float end_deg)
{
   float sr = begin_deg * .0174533, er = end_deg * .0174533;
   float dr = (v_cnt > 1) ? (er - sr)/(v_cnt - 1) : (er - sr)/2;
   unsigned n = 0;

   if (!v_arr || !v_cnt)
      return;

   if (v_cnt == 1) {
      v_arr[0].x = .5 * cos(dr + sr);
      v_arr[0].y = .5 * sin(dr + sr);
      return;
   }

   while (n < v_cnt) {
      v_arr[n].x = .5 * cos(n * dr + sr);
      v_arr[n].y = .5 * sin(n * dr + sr);
      n++;
   }
}


void gluf_tessellate_2d(vec2 **v_arr, unsigned *v_cnt, unsigned loop)
{
   unsigned n = 0, final;
   vec2 *ret;

   if (!v_arr || !*v_arr || !v_cnt || *v_cnt < 2)
      return;

   final = (*v_cnt - 1) * 3;
   if (loop)
      final += 3;

   ret = (vec2*)calloc(final, sizeof(vec2));
   if (!ret)
      return;

   while (n < *v_cnt - 1) {
      ret[3 * n]     = (*v_arr)[n];
      ret[3 * n + 2] = (*v_arr)[n + 1];
      n++;
   }

   if (loop) {
      ret[3 * n] = (*v_arr)[*v_cnt - 1];
      ret[3 * n + 2] = (*v_arr)[0];
   }

   free(*v_arr);
   *v_arr = ret;
   *v_cnt = final;
}

/* Assumes vertex coordinates  x = {-.5,.5}, y = {-.5,.5} */
void gluf_tex_coords_2d(const vec2 *v_arr, vec2 *t_arr, unsigned v_cnt)
{
   unsigned n = 0;
   float mx_x = 0, mn_x = 0, mx_y = 0, mn_y = 0;

   if (!v_arr || !t_arr)
      return;

   /* Copy and find  x/y minimum/maximum... */
   while (n < v_cnt) {
      t_arr[n].x = v_arr[n].x;
      t_arr[n].y = v_arr[n].y;

      if (v_arr[n].x > mx_x)
         mx_x = v_arr[n].x;
      if (v_arr[n].x < mn_x)
         mn_x = v_arr[n].x;
      if (v_arr[n].y > mx_y)
         mx_y = v_arr[n].y;
      if (v_arr[n].y < mn_y)
         mn_y = v_arr[n].y;
      n++;
   }

   gluf_translate_2d(t_arr,v_cnt,-mn_x,-mn_y);
   mx_x -= mn_x;
   mx_y -= mn_y;
   gluf_scale_2d(t_arr,v_cnt,(mx_x) ? 1/mx_x : 0,(mx_y) ? 1/mx_y : 0);
}

void gluf_scale_2d(vec2 *v_arr,unsigned v_cnt, float scale_x, float scale_y)
{
   unsigned n = 0;

   if (!v_arr || scale_x == 1 && scale_y == 1)
      return;

   while (n < v_cnt) {
     v_arr[n].x *= scale_x;
     v_arr[n].y *= scale_y;
      n++;
   }
}

void gluf_translate_2d(vec2 *v_arr, unsigned v_cnt, float x, float y)
{
   unsigned n = 0;

   if (!v_arr || !x && !y)
      return;

   while (n < v_cnt) {
      v_arr[n].x += x;
      v_arr[n].y += y;
      n++;
   }
}

void gluf_rotate_2d(vec2 *v_arr, unsigned v_cnt, float deg)
{
   unsigned n = 0;
   float r, *x, *y, theta, in_rad = deg * .0174533;

   if (!v_arr || !deg)
      return;

   while (n < v_cnt) {
      x     = &v_arr[n].x;
      y     = &v_arr[n].y;
      r     = sqrt(*x * *x + *y * *y);
      theta = atan2(*y,*x) + in_rad;
      *x    = r * cos(theta);
      *y    = r * sin(theta);
      n++;
   }
}

void gluf_transform(vec2 *v_arr, unsigned v_cnt, mat33 *m)
{
   unsigned n = 0;

   if (!v_arr || !m)
      return;

   while (n < v_cnt)
      vec2_transform33(&v_arr[n++],m);
}


gluf_bitmap *gluf_get_bitmap(const char *file)
{
   FREE_IMAGE_FORMAT fmt;
   FIBITMAP *fi_data;
   gluf_bitmap *ret = 0;
   unsigned size;

   #ifdef USING_FREEIMAGE_RUNTIME_LD
      static int lib_loaded = -1;
      if (lib_loaded < 0)
         lib_loaded = ext_load_freeimage();
      if (!lib_loaded)
         return 0;
   #endif

   if (!file)
      return 0;

   fmt = FreeImage_GetFileType(file, 0);
   if (fmt == FIF_UNKNOWN)
      fmt = FreeImage_GetFIFFromFilename(file);
   if (fmt != FIF_UNKNOWN) {
      if (FreeImage_FIFSupportsReading(fmt))
         fi_data = FreeImage_Load(fmt,file,0);
      if (fi_data) {
         ret = (gluf_bitmap*)malloc(sizeof(gluf_bitmap));

         if (ret) {
            ret->w    = FreeImage_GetWidth(fi_data);
            ret->h    = FreeImage_GetHeight(fi_data);
            ret->bpp  = FreeImage_GetBPP(fi_data);
            size      = ret->w * ret->h * ret->bpp/8.0;
            ret->data = (unsigned char*)malloc(size * sizeof(char));

            if (ret->data) {
               memcpy(ret->data, FreeImage_GetBits(fi_data),size);
            }
         }
         FreeImage_Unload(fi_data);
      }
      else printf("Could not read %s...\n",file);
   }
   else printf("Could not load %s...\n",file);

   return ret;
}

void gluf_get_bitmap_async(const char *file, gluf_bitmap **location)
{
   async_param *p;
   void *id;

   if (!file || !location)
      return;

   p = (async_param*)malloc(sizeof(async_param));
   if (p) {
      p->image_file = strdup(file);
      p->loc_ptr    = location;

      platform_new_thread(get_bm_async,p,&id);
   }
}

static t_func_t get_bm_async(void *param)
{
   async_param *p = (async_param*)param;
   gluf_bitmap *bm;

   if (p) {
      bm = gluf_get_bitmap(p->image_file);

      if (bm)
         *(p->loc_ptr) = bm;

      if (p->image_file)
         free(p->image_file);
      free(p);
   }

   platform_end_thread(0);
}

gluf_bitmap *gluf_empty_bitmap(unsigned width,unsigned height,unsigned bpp)
{
   gluf_bitmap *obj = (gluf_bitmap*)malloc(sizeof(gluf_bitmap));
   unsigned size;

   if (obj) {
      obj->w    = width;
      obj->h    = height;
      obj->bpp  = bpp;
      size      = obj->w * obj->h * obj->bpp/8.0;
      obj->data = (unsigned char*)calloc(size,sizeof(char));
   }
   return obj;
}

void gluf_bitmap_flip_vertical(gluf_bitmap *bm)
{
   unsigned char *temp, *bm_n, *bm_m;
   unsigned n = 0, m, stride;

   if (!bm || !bm->w || !bm->h || !bm->bpp || !bm->data)
      return;

   m      = bm->h - 1;
   stride = bm->w * bm->bpp/8 * sizeof(char);
   temp   = (unsigned char*)malloc(stride);

   if (!temp)
      return;

   while (n < m) {
      bm_n = bm->data + stride * n++;
      bm_m = bm->data + stride * m--;

      memcpy(temp,bm_n,stride);
      memcpy(bm_n,bm_m,stride);
      memcpy(bm_m,temp,stride);
   }
   free(temp);
}


void gluf_bitmap_flip_horizontal(gluf_bitmap *bm)
{
   unsigned char *temp, *ptr;
   unsigned n, m, row = 0, height;
   unsigned stride;

   if (!bm || !bm->w || !bm->h || !bm->bpp || !bm->data)
      return;

   height = bm->h;
   stride = bm->bpp/8 * sizeof(char);
   temp   = (unsigned char*)malloc(stride);

   if (!temp)
      return;

   ptr = bm->data;

   while (row < height) {
      n = 0;
      m = bm->w - 1;
      while (n < m) {
         memcpy(temp,ptr + stride * n,stride);
         memcpy(ptr + stride * n,ptr + stride * m,stride);
         memcpy(ptr + stride * m,temp,stride);
         n++;
         m--;
      }
      row++;
      ptr += bm->w * bm->bpp/8;
   }

   free(temp);
}

#define THUMB_SZ 256

gluf_bitmap *gluf_bitmap_extract_thumb(gluf_bitmap *bm)
{
   gluf_bitmap *ret;
   char *ptr, *sub_ptr;
   unsigned n = 0, m;

   if (!bm || !bm->w || !bm->h || !bm->bpp || !bm->data)
      return 0;

   if (bm->w <= THUMB_SZ || bm->h <= THUMB_SZ)
      return 0;

   ret = gluf_empty_bitmap(THUMB_SZ,THUMB_SZ,bm->bpp);

   if (ret) {
      sub_ptr = ret->data;
      ptr = bm->data + (bm->w * bm->h/2 * bm->bpp/8) + (bm->w/2 * bm->bpp/8);
      ptr -= (THUMB_SZ/2 * bm->w * bm->bpp/8) + (THUMB_SZ/2 * bm->bpp/8);

      while (n < THUMB_SZ) {
         memcpy(sub_ptr,ptr,THUMB_SZ * bm->bpp/8);
         n++;
         ptr += bm->w * bm->bpp/8;
         sub_ptr += THUMB_SZ * bm->bpp/8;
      }

      return ret;
   }

   return 0;
}

void gluf_bitmap_to_file(gluf_bitmap *bm, const char *file)
{
   FREE_IMAGE_FORMAT fmt;
   FIBITMAP *fi_data = 0;

   #ifdef USING_FREEIMAGE_RUNTIME_LD
      static int lib_loaded = -1;
      if (lib_loaded < 0)
         lib_loaded = ext_load_freeimage();
      if (!lib_loaded) {
         printf("Oh no.. ");
         return;
      }
   #endif

   if (!bm || !file) {
      printf("Missing paramater(s)...\n");
      return;
   }

   fmt = FreeImage_GetFIFFromFilename(file);

   if (fmt != FIF_UNKNOWN) {

      if (FreeImage_FIFSupportsWriting(fmt))
         fi_data = FreeImage_ConvertFromRawBits(
            (BYTE*)bm->data,
            bm->w,
            bm->h,
            bm->w * bm->bpp/8,
            bm->bpp,
            FI_RGBA_RED_MASK,
            FI_RGBA_GREEN_MASK,
            FI_RGBA_BLUE_MASK,
            0);

      if (fi_data) {
         if (!FreeImage_Save(fmt,fi_data,file,0))
            printf("Did not save %s...\n",file);

         FreeImage_Unload(fi_data);
      }
      else
         printf("Could not save %s...\n",file);
   }
   else
      printf("Unknown format for %s...\n",file);

}

/* void gluf_bitmap_draw(gluf_bitmap *s,gluf_bitmap *d,vec2 *sa, vec2 *da, unsigned cnt)
{
   unsigned src_x, src_y, dst_x, dst_y, lmt_x, lmt_y, clr_cmp, Bpp, n = 0;

   if (!s || !d || !sa || !da)
      return;

   if (s->bpp != d->bpp) {
      printf ("Bitmap transfer: dissimilar pixel formats...done.\n");
      return;
   }

   Bpp = s->bpp/8;

   while (n < cnt) {
      src_y = sa[n].y;
      dst_y = da[n].y;
      lmt_y = sa[n + 1].y;

      while (src_y <= lmt_y) {
         src_x = sa[n].x;
         dst_x = da[n].x;
         lmt_x = sa[n + 2].x;

         while (src_x <= lmt_x) {
            clr_cmp = 0;

            while (clr_cmp < Bpp) {
               d->data[d->w * dst_y + dst_x + clr_cmp]
                  += s->data[s->w * src_y + src_x + clr_cmp];
               clr_cmp++;
            }
            src_x++;
            dst_x++;
         }
         src_y++;
         dst_y++;
      }
      n += 4;
   }
} */


void gluf_erase_bitmap(gluf_bitmap *bm)
{
   if (!bm)
      return;

   if (bm->data)
      free(bm->data);

   free(bm);
}



#ifdef __cplusplus
}
#endif

