#ifndef __GLUF_H__
#define __GLUF_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "types.h"
#include "matrix44.h"

//typedef  vec2            coord_2d;
typedef  vec4           area_2d;
typedef  vec3           coord_3d;
typedef  BITMAP_T       gluf_bitmap;
typedef  SIMPLE_MAT_T   gluf_mat;
typedef union { struct { unsigned char r, g, b, a;}; unsigned int i; } gluf_pixel;

typedef struct {
   unsigned x, y, sz, lz4_sz;
   char *data;
}
gluf_rle_data;

/* Bitmap container... */
gluf_bitmap *gluf_get_bitmap(const char *file);
void gluf_get_bitmap_async(const char *file, gluf_bitmap **location);
gluf_bitmap *gluf_empty_bitmap(unsigned width,unsigned height, unsigned bpp);
void gluf_bitmap_flip_vertical(gluf_bitmap *bm);
void gluf_bitmap_flip_horizontal(gluf_bitmap *bm);
gluf_bitmap *gluf_bitmap_extract_thumb(gluf_bitmap *bm);
void gluf_bitmap_to_file(gluf_bitmap *bm, const char *file);
void gluf_erase_bitmap(gluf_bitmap *bm);

/* Alpha information is lost:
   - expects input in xxxa format, returns xxx1 format.
   - used for buffer capture disk dumps which return bgra
     format with no relevant alpha information... */

void gluf_image_mix_in(char *dest, char *src, int x, int y, unsigned order);
void gluf_image_encode(gluf_rle_data *rd);
void gluf_image_encode_fast_buffer_size(unsigned sz);
gluf_rle_data *gluf_image_encode_fast(char *data,int x, int y);
void gluf_image_decode(gluf_rle_data *rd);
void gluf_image_decode_invert(gluf_rle_data *rd);
void gluf_free_rle(gluf_rle_data *rle);

/* Two dimesional coordinate array functions... */
void gluf_arc_2d(vec2 *v_arr, unsigned v_cnt, float begin_deg, float end_deg);
/* Assumes a perimeter defining a polygon around origin. Array is resized
to reflect final triangle vertex count. */
void gluf_tessellate_2d(vec2 **v_arr, unsigned *v_cnt, unsigned loop);
void gluf_tex_coords_2d(const vec2 *v_arr, vec2 *t_arr, unsigned v_cnt);
void gluf_scale_2d(vec2 *v_arr,unsigned v_cnt, float scale_x, float scale_y);
void gluf_translate_2d(vec2 *v_arr, unsigned v_cnt, float x, float y);
void gluf_rotate_2d(vec2 *v_arr, unsigned v_cnt, float deg);
void gluf_transform(vec2 *v_arr, unsigned v_cnt, mat33 *matrix);

#ifdef __cplusplus
}
#endif

#endif /* __GLUF_H__ */

