#ifdef __cplusplus
extern "C" {
#endif

#define USING_OPENGL
#include "platform.h"
#include <GL/glu.h>
#include "renderer_api.h"

#include "file_parser.h"
#include "hotspot.h"

#define MAX_SPOTS 100

typedef struct {
   IMPLEMENTS_UI_C3D_HOTSPOT
   void (*save_erase)(void *obj);

   unsigned active_id;
   c3d_element *active_obj;
   float active_for;

   ui_element *hdr[MAX_SPOTS + 1];
   ui_element *msg[MAX_SPOTS + 1];
   ui_element *pic[MAX_SPOTS + 1];
   ui_element *top_panel;
   ui_element *bottom_panel;
   unsigned disable_visual;

   char *nar[MAX_SPOTS + 1];
   stream_src *narrator;
   float audio_level;
   unsigned started_narrator;
   unsigned disable_audio;

}
hotspot;


static void erase(void *obj);
static void hs_set(c3d_ui_hotspot *h,hotspot_directive what,float val);
static int compile_object(parse_list *p, hotspot *h);

static void on_hit(void* self,c3d_element* who);
static void no_hit(void* self,c3d_element* who);
static void active_update(void* self,c3d_update* info);
static void all_display(void *self);
static void sphere_list(void *owner,unsigned id);

static unsigned initialized = 0;
static unsigned sphere_list_id;
static color_t clr1 = {.95,.95,.5,.7};
static color_t clr2 = {.65,.95,.65,.7};

c3d_ui_hotspot *create_c3d_ui_hotspot(const char *src, unsigned m)
{
   c3d_domain *d;
   ui_container *c;
   hotspot *h;
   parse_list *pl;

   if (!src)
      return 0;

   h = (hotspot*)calloc(1,sizeof(hotspot));
   if (!h)
      return 0;

   c = create_container();
   if (!c) {
      free(h);
      return 0;
   }

   memcpy(h,c,sizeof(ui_container));
   free(c);
   c = (ui_container*)h;

   h->save_erase = c->erase;
   c->erase = &erase;

   d = c3d_create_domain();
   if (!d) {
      free(h);
      return 0;
   }
   h->spots = d;

   if (!(pl = file_parser_get_objects(src,m)) || !compile_object(pl,h)) {
      h->erase(h);
      return 0;
   }

   file_parser_erase_list(pl);

   h->top_panel = (ui_element*)create_custom_icon(.02);
   h->top_panel->place(h->top_panel,0,.96,1,.04);
   h->top_panel->look = material_is(0,0,0,.85,0);
   h->bottom_panel = (ui_element*)create_custom_icon(.02);
   h->bottom_panel->place(h->bottom_panel,0,0,1,.2);
   h->bottom_panel->look = material_is(0,0,0,.85,0);

   h->set = &hs_set;
   h->audio_level = .5;

   if (!initialized) {
      sphere_list_id = renderer_register_list_function(h,&sphere_list);
      initialized = 1;
   }

   return (c3d_ui_hotspot*)h;
}

static int compile_object(parse_list *pl, hotspot *h)
{
   unsigned id = 1;

   parse_obj *obj;
   ui_element *e;
   ui_text *tx;
   c3d_element *c;
   loc_list *loc_ptr;

   if (!h)
      return 0;

   while (pl) {

      obj = pl->p;
      if (!obj) {
         id++;
         pl = pl->more;
         continue;
      }

      if (obj->hdr) {
         tx = create_text(80,obj->hdr);
         tx->font = ui_pretty_fnt;
         tx->justify_top = 1;
         tx->centered = 1;
         e = to_static_text(tx);
         tx = 0;
         obj->hdr = 0;

         if (obj->hdr_clr.a) {
            decor_all_colors(
               (ui_decor*)&e->look,
               obj->hdr_clr.r,obj->hdr_clr.g,obj->hdr_clr.b,obj->hdr_clr.a);
         }
         else {
            decor_all_colors(
               (ui_decor*)&e->look,
               clr2.r,clr2.g,clr2.b,clr2.a);
         }

         e->place(e,0,0,1,1);

         h->hdr[id] = e;
         e = 0;
      }

      if (obj->msg) {
         tx = create_text(90,obj->msg);
         tx->font = ui_pretty_fnt;
         tx->centered = 1;
         e = to_static_text(tx);
         tx = 0;
         obj->msg = 0;

         if (obj->msg_clr.a) {
            decor_all_colors(
               (ui_decor*)&e->look,
               obj->msg_clr.r,obj->msg_clr.g,obj->msg_clr.b,obj->msg_clr.a);
         }
         else {
            decor_all_colors(
               (ui_decor*)&e->look,
               clr1.r,clr1.g,clr1.b,clr1.a);
         }

         h->msg[id] = e;
         e = 0;
      }

      if (obj->nar) {
         h->nar[id] = obj->nar;
         obj->nar = 0;
      }

      if (obj->pic) {
         e = (ui_element*)create_custom_icon(.1);
         decor_all_materials(
            (ui_decor*)&e->look,
            .9,.9,.9,.9,
            renderer_register_texture_file(obj->pic));
         e->place(e,.025,.205,.2,.2);
         h->pic[id] = e;
         e = 0;
      }

      loc_ptr = obj->at;
      while (loc_ptr) {

         c = c3d_create_element(in_volume(
            loc_ptr->at.v.x,loc_ptr->at.v.y,loc_ptr->at.v.z,loc_ptr->at.v.r));
         c->id      = id;
         c->owner   = (void*)h;
         c->on_hit  = &on_hit;
         c->no_hit  = &no_hit;
         c->display = &all_display;
         h->spots->add_object(h->spots,c);
         h->spots->add_collidable(h->spots,c);
         c = 0;

         loc_ptr = loc_ptr->more;
      }

      id++;
      pl = pl->more;
   }

   return 1;
}

static void sphere_list(void *owner,unsigned id)
{
   GLUquadricObj* quadRender= gluNewQuadric();

   gluQuadricDrawStyle(quadRender,GLU_FILL);
   gluQuadricOrientation(quadRender,GLU_OUTSIDE);
   gluQuadricNormals(quadRender,GLU_SMOOTH);
   gluSphere(quadRender,1,15,15);

   gluDeleteQuadric(quadRender);
}

static void erase(void *obj)
{
   hotspot *h = (hotspot*)obj;
   ui_container *uc = (ui_container*)h;
   unsigned n = 1;

   if (!h)
      return;

   if (h->spots)
      h->spots->erase(h->spots);

   uc->detach_all(uc);

   erase_stream_source(h->narrator);

   while (n < MAX_SPOTS + 1) {
      if (h->hdr[n])
         h->hdr[n]->erase(h->hdr[n]);
      if (h->msg[n])
         h->msg[n]->erase(h->msg[n]);
      if (h->pic[n])
         h->pic[n]->erase(h->pic[n]);
      if (h->nar[n])
         free(h->nar[n]);
      n++;
   }

   if (h->top_panel)
       free(h->top_panel);
   if (h->bottom_panel)
      free(h->bottom_panel);

   h->save_erase(h);
}

static void hs_set(c3d_ui_hotspot *obj, hotspot_directive what,float val)
{
   hotspot *h = (hotspot*) obj;
   ui_container *uc = (ui_container*)h;

   if (!h)
      return;

   switch (what) {

      case HS_USE_AUDIO:
         if (val) {
            h->disable_audio = 0;
            if (h->active_id) {
               h->started_narrator = 1;
               h->narrator = create_stream_source(h->nar[h->active_id]);
               if (h->narrator) {
                  alSourcef(h->narrator->src,AL_GAIN,h->audio_level);
                  alSourcePlay(h->narrator->src);
               }
            }
         }
         else {
            h->disable_audio = 1;
            h->started_narrator = 0;
            erase_stream_source(h->narrator);
            h->narrator = 0;
         }
         break;

      case HS_SHOW_VISUAL:
         if (val) {
            h->disable_visual = 0;
            if (h->active_id) {
               uc->add_item(uc,h->top_panel);
               uc->add_item(uc,h->bottom_panel);
               uc->add_item(uc,h->hdr[h->active_id]);
               uc->add_item(uc,h->msg[h->active_id]);
               uc->add_item(uc,h->pic[h->active_id]);
            }
         }
         else {
            h->disable_visual = 1;
            uc->detach_all(uc);
         }
         break;

      case HS_SUSPEND:
         if (val) {
            erase_stream_source(h->narrator);
            h->started_narrator = 0;
            h->narrator = 0;
         }
         else {
            if (h->active_id && !h->disable_audio && !h->started_narrator) {
               h->started_narrator = 1;
               h->narrator = create_stream_source(h->nar[h->active_id]);
               if (h->narrator) {
                  alSourcef(h->narrator->src,AL_GAIN,h->audio_level);
                  alSourcePlay(h->narrator->src);
               }
            }
         }
         break;

      case HS_AUDIO_LEVEL:
         h->audio_level = val;
         if (h->narrator)
            alSourcef(h->narrator->src,AL_GAIN,val);
         break;
   }
}

static void on_hit(void* self,c3d_element* who)
{
   c3d_element *c = (c3d_element*)self;
   ui_container *uc;
   hotspot *h;

   if (!c)
      return;

   h  = (hotspot*)c->owner;
   uc = (ui_container*)h;

   if (h->active_id < c->id) {

      uc->detach_all(uc);
      erase_stream_source(h->narrator);

      if (!h->disable_visual) {
         uc->add_item(uc,h->top_panel);
         uc->add_item(uc,h->bottom_panel);
         uc->add_item(uc,h->hdr[c->id]);
         uc->add_item(uc,h->msg[c->id]);
         uc->add_item(uc,h->pic[c->id]);
      }

      c->update           = &active_update;
      h->narrator         = 0;
      h->active_id        = c->id;
      h->active_obj       = c;
      h->active_for       = 0;
      h->started_narrator = 0;
   }
}

static void no_hit(void* self,c3d_element* who)
{
   c3d_element *c = (c3d_element*)self;
   ui_container *uc;
   hotspot *h;

   if (!c)
      return;

   h  = (hotspot*)c->owner;
   uc = (ui_container*)h;

   if (h->active_obj == c) {

      uc->detach_all(uc);
      erase_stream_source(h->narrator);

      c->update           = 0;
      h->narrator         = 0;
      h->active_obj       = 0;
      h->active_id        = 0;
      h->active_for       = 0;
      h->started_narrator = 0;
   }
}

static void active_update(void* self,c3d_update* info)
{
   c3d_element *c = (c3d_element*)self;
   hotspot *h;

   if (!c)
      return;

   h = (hotspot*)c->owner;

   if (h->active_obj == c) {

      h->active_for += info->dt;

      if (!h->disable_audio && !h->started_narrator && h->active_for > .5) {
         h->started_narrator = 1;
         h->narrator = create_stream_source(h->nar[c->id]);
         if (h->narrator) {
            alSourcef(h->narrator->src,AL_GAIN,h->audio_level);
            alSourcePlay(h->narrator->src);
         }
      }
   }
}

static void all_display(void *self)
{
   c3d_element *c = (c3d_element*)self;
   hotspot *h;
   float r,g,b;

   if (!c)
      return;

   h = (hotspot*)c->owner;

   r = c->id%2 * c->id%10/(float)10;
   g = c->id%3 * c->id%10/(float)10;
   b = (!(c->id%2) && !(c->id%3)) * c->id%10/(float)10;

   glEnable(GL_CULL_FACE);
   glEnable(GL_COLOR_MATERIAL);
   glDisable(GL_TEXTURE_2D);
   glColor4f(r,g,b,.5);

   glTranslatef(c->at.x,c->at.y,c->at.z);
   glScalef(c->at.r,c->at.r,c->at.r);
   glCallList(sphere_list_id);
   glScalef(1/c->at.r,1/c->at.r,1/c->at.r);
   glTranslatef(-c->at.x,-c->at.y,-c->at.z);
   glDisable(GL_CULL_FACE);

}



#ifdef __cplusplus
}
#endif
