#include "gl_extensions.h"
#include "line_shader.h"


static void enumerate(shader_program p);
static void execute(shader_program p);


line_shader_t line_shader = {

	INIT_MODULE_NAME("line_shader")

   INIT_VERTEX_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "attribute vec3 av3vertex;"
      "attribute vec4 av4color;"

      "uniform bool use_color_vertex;"
      "uniform mat4 um44mvp;"
      "uniform vec4 uv4color;"

      "varying vec4 vv4color;"


      "void main() {"
         "vv4color    = (use_color_vertex) ? av4color : uv4color;"
         "gl_Position = um44mvp * vec4(av3vertex,1.0);"
      "}"

   INIT_FRAGMENT_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "varying vec4 vv4color;"

      "void main(void) {"
         "gl_FragColor = vv4color;"
      "}"

   INIT_ALLOCATED_SOURCES(0)
   INIT_STD_COMPILE_FUNCTION
   INIT_ENUMERATE_FUNCTION(&enumerate)
   INIT_EXECUTE_FUNCTION(&execute)
   INIT_SHADER_API_BLOCK_END
};

static GLint loc[5];

static void enumerate(shader_program p)
{
   line_shader_t *s = (line_shader_t*)p;

   if (!s || !glIsProgram(s->program))
      return;

   loc[0] = glGetUniformLocation(s->program,"um44mvp");
   loc[1] = glGetUniformLocation(s->program,"uv4color");
   loc[2] = glGetAttribLocation(s->program,"av3vertex");
   loc[3] = glGetAttribLocation(s->program,"av4color");
   loc[4] = glGetUniformLocation(s->program,"use_color_vertex");
}

static void execute(shader_program p)
{
   line_shader_t *s = (line_shader_t*)p;

   if (!s)
      return;

   if (glIsProgram(s->program))
      glUseProgram(s->program);
   else
      return;

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,s->vbo);

   if (!s->depth_mask)
      glDepthMask(GL_FALSE);

   glLineWidth(s->line_width);

   glUniformMatrix4fv(loc[0],1,GL_FALSE,(float*)s->um44mvp);

   glVertexAttribPointer(loc[2],3,GL_FLOAT,GL_FALSE,0,s->av3vertex);
   glEnableVertexAttribArray(loc[2]);

   if (s->av4color) {
      glUniform1i(loc[4],1);
      glVertexAttribPointer(loc[3],4,GL_FLOAT,GL_FALSE,0,s->av4color);
      glEnableVertexAttribArray(loc[3]);
   }
   else {
      glUniform1i(loc[4],0);
      glUniform4fv(loc[1],1,s->uv4color);
   }

   glDrawArrays((s->loop) ? GL_LINE_LOOP : GL_LINE_STRIP,0,s->count);

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,0);
   glDisableVertexAttribArray(loc[2]);
   glDisableVertexAttribArray(loc[3]);

   glUseProgram(0);
   glDepthMask(GL_TRUE);
}
