#include "gl_extensions.h"
#include "model_shader_nicest.h"


static void enumerate(shader_program p);
static void execute(shader_program p);


model_shader_nicest_t model_shader_nicest = {

   INIT_MODULE_NAME("model_shader_nicest")

   INIT_VERTEX_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "attribute vec3 av3vertex;"
      "attribute vec3 av3normal;"
      "attribute vec2 av2texture;"

      "uniform mat4 um44mvp;"
      "uniform mat4 um44mv;"
      "uniform vec4 Lp;"

      "varying vec2 vv2tex_coord;"
      "varying vec3 vv3normal;"
      "varying vec3 vv3eye;"
      "varying float mag;"

      "void main() {"
         "vv3normal    = normalize(um44mv * vec4(av3normal,0)).xyz;"
         "vv3eye       = -(um44mv * vec4(av3vertex,1.0)).xyz;"
         "mag          = length(Lp.xyz - vv3eye);"
         "vv2tex_coord = av2texture;"
         "gl_Position  = um44mvp * vec4(av3vertex,1.0);"
      "}"

   INIT_FRAGMENT_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "uniform vec4 Ka;"
      "uniform vec4 Kd;"
      "uniform vec4 Ks;"
      "uniform vec4 Ke;"
      "uniform float Ns;"

      "uniform vec4 La;"
      "uniform vec4 Ld;"
      "uniform vec4 Ls;"
      "uniform vec4 Lp;"
      "uniform float cutoff;"
      "uniform float floor;"

      "uniform sampler2D texture1;"
      "uniform int have_texture;"

      "varying vec2 vv2tex_coord;"
      "varying vec3 vv3normal;"
      "varying vec3 vv3eye;"
      "varying float mag;"

      "void main(void) {"

         "vec4 color = vec4(1.0);"
         "vec3 low = vec3(0.0);"
         "vec3 high = vec3(1.0);"

         "if (have_texture > 0)"
            "color = texture2D(texture1,vv2tex_coord);"

         "vec3 eye = vv3eye;"
         "vec3 L   = normalize(eye + Lp.xyz);"
         "eye      = normalize(eye);"
         "vec3 h   = normalize(L + eye);"
         "vec3 n   = normalize(vv3normal);"

         "float coef_d = max(dot(L,n),0.0);"
         "float coef_s = max(pow(dot(h,n),Ns),0.0);"
         "float i = clamp(1.0 - ((cutoff > 0.0) ? mag/cutoff : 0.0),floor,1.0);"

         "vec3 rgb ="
            "color.rgb * Ke.rgb +"
            "color.rgb * Ka.rgb * La.rgb +"
            "color.rgb * coef_d * Ld.rgb * Kd.rgb +"
            "coef_s * Ls.rgb * Ks.rgb;"

         "rgb      = clamp(i * rgb,low,high);"
         "float a  = Ke.a * Ka.a * Kd.a * Ks.a * color.a;"
         "gl_FragColor = vec4(rgb,a);"
      "}"

   INIT_ALLOCATED_SOURCES(0)
   INIT_STD_COMPILE_FUNCTION
   INIT_ENUMERATE_FUNCTION(&enumerate)
   INIT_EXECUTE_FUNCTION(&execute)
   INIT_SHADER_API_BLOCK_END
};

static GLint loc[18];

static void enumerate(shader_program p)
{
   model_shader_nicest_t *s = (model_shader_nicest_t*)p;

   if (!s || !glIsProgram(s->program))
      return;

   loc[0] = glGetAttribLocation(s->program,"av3vertex");
   loc[1] = glGetAttribLocation(s->program,"av3normal");
   loc[2] = glGetAttribLocation(s->program,"av2texture");

   loc[3] = glGetUniformLocation(s->program,"um44mvp");
   loc[4] = glGetUniformLocation(s->program,"um44mv");

   loc[5] = glGetUniformLocation(s->program,"Ka");
   loc[6] = glGetUniformLocation(s->program,"Kd");
   loc[7] = glGetUniformLocation(s->program,"Ks");
   loc[8] = glGetUniformLocation(s->program,"Ke");
   loc[9] = glGetUniformLocation(s->program,"Ns");

   loc[10] = glGetUniformLocation(s->program,"La");
   loc[11] = glGetUniformLocation(s->program,"Ld");
   loc[12] = glGetUniformLocation(s->program,"Ls");
   loc[13] = glGetUniformLocation(s->program,"Lp");
   loc[14] = glGetUniformLocation(s->program,"cutoff");
   loc[15] = glGetUniformLocation(s->program,"floor");

   loc[16] = glGetUniformLocation(s->program,"texture1");
   loc[17] = glGetUniformLocation(s->program,"have_texture");


}

static void execute(shader_program p)
{
   model_shader_nicest_t *s = (model_shader_nicest_t*)p;
   geom_part_t *g;
   material_t  m;
   unsigned n = 0;

   if (!s || !s->um44mvp || !s->um44mv || !s->list)
      return;

   if (glIsProgram(s->program))
      glUseProgram(s->program);
   else
      return;

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,s->vbo);

   glEnableVertexAttribArray(loc[0]);
   glEnableVertexAttribArray(loc[1]);
   glEnableVertexAttribArray(loc[2]);

   glUniformMatrix4fv(loc[3],1,GL_FALSE,(float*)s->um44mvp);
   glUniformMatrix4fv(loc[4],1,GL_FALSE,(float*)s->um44mv);

   glUniform4fv(loc[10],1,s->list->light.La.fv);
   glUniform4fv(loc[11],1,s->list->light.Ld.fv);
   glUniform4fv(loc[12],1,s->list->light.Ls.fv);
   glUniform4fv(loc[13],1,s->list->light.Lp.fv);
   glUniform1fv(loc[14],1,&s->list->light.cutoff_distance);
   glUniform1fv(loc[15],1,&s->list->light.minimum_level);


   glActiveTexture(GL_TEXTURE0);
   glUniform1i(loc[16],0);

   while (n < s->list->part_count) {

      g = &s->list->part[n];
      m = g->material;

      if (g->vertex_count) {

         if (m.map && glIsTexture(m.map)) {
            glUniform1i(loc[17],1);
            glBindTexture(GL_TEXTURE_2D,m.map);
         }
         else
            glUniform1i(loc[17],0);

         glUniform4fv(loc[5],1,m.Ka.fv);
         glUniform4fv(loc[6],1,m.Kd.fv);
         glUniform4fv(loc[7],1,m.Ks.fv);
         glUniform4fv(loc[8],1,m.Ke.fv);
         glUniform1fv(loc[9],1,&m.Ns);
         //printf("Ns: %f\n",m.Ns);

         glVertexAttribPointer(
            loc[0],3,GL_FLOAT,GL_FALSE,0,(void*)g->vertex_start);

         if (g->normal_start) {
            glVertexAttribPointer(
               loc[1],3,GL_FLOAT,GL_FALSE,0,(void*)g->normal_start);
         }
         else {
            glVertexAttribPointer(
               loc[1],3,GL_FLOAT,GL_FALSE,0,(void*)g->vertex_start);
         }
         if (g->texture_start) {
            glVertexAttribPointer(
               loc[2],2,GL_FLOAT,GL_FALSE,0,(void*)g->texture_start);
         }
         else {
            glVertexAttribPointer(
               loc[2],2,GL_FLOAT,GL_FALSE,0,(void*)g->vertex_start);
         }

         glDrawArrays(GL_TRIANGLES,0,g->vertex_count);
      }
      n++;
   }

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,0);
   glDisableVertexAttribArray(loc[0]);
   glDisableVertexAttribArray(loc[1]);
   glDisableVertexAttribArray(loc[2]);
   glUseProgram(0);
   glBindTexture(GL_TEXTURE_2D,0);
}
