#ifdef __cplusplus
extern "C" {
#endif

#include "platform.h"

/* Note: this is kind of nasty: no real commonality exists between the two
differing platform precision time functions... */

static unsigned initialized = 0;
static unsigned use_high_precision;
static unsigned long backup_last_time;
static unsigned long backup_now;
static double ret_val;

#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)

static LARGE_INTEGER last_time;
static LARGE_INTEGER now;
static LARGE_INTEGER freq;

double platform_delta_timer(void)
{
   if (!initialized) {
      use_high_precision = QueryPerformanceFrequency(&freq);
      if (use_high_precision)
         QueryPerformanceCounter(&last_time);
      else backup_last_time = clock();
      initialized = 1;
   }

   if (use_high_precision) {
      QueryPerformanceCounter(&now);
      ret_val = (now.QuadPart - last_time.QuadPart)/(double)freq.QuadPart;
      last_time = now;
   }
   else {
      backup_now = clock();
      ret_val = (backup_now - backup_last_time)/(double)CLOCKS_PER_SEC;
      backup_last_time = backup_now;
   }
   return ret_val;
}

#else /* POSIX */

static struct timeval last_time;
static struct timeval now;

double platform_delta_timer(void)
{
   if (!initialized) {
      use_high_precision = !gettimeofday(&last_time,0);
      initialized = 1;
   }

   if (use_high_precision) {
      if (!gettimeofday(&now,0)) {
         ret_val =
            ((now.tv_sec - last_time.tv_sec) * CLOCKS_PER_SEC
            + now.tv_usec - last_time.tv_usec)/(double) CLOCKS_PER_SEC;
         last_time = now;
      }
   }
   else {
     backup_now = clock();
     ret_val = (backup_now - backup_last_time)/(double)CLOCKS_PER_SEC;

     /* Sometimes Linux systems return a sub-second value less than the previous
     one--and no, it's not just unsigned type rollover... */
     if  (ret_val > 2000)
         ret_val = (backup_last_time - backup_now)/(double)CLOCKS_PER_SEC;

     backup_last_time = backup_now;
  }

  return ret_val;
}

#endif

#ifdef __cplusplus
}
#endif
