#ifndef __PLATFORM_H__
#define __PLATFORM_H__

/* For Windows stdlib -- must be defined BEFORE inclusion. */
#ifndef _CRT_SECURE_NO_WARNINGS
   #define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>

#if defined(WIN32) || defined(_WIN32) || defined(WIN64) || defined(_WIN64)

   #define WIN32_LEAN_AND_MEAN
   #define _USE_MSVC
   #include <windows.h>
   #include <process.h>
   #define snprintf        _snprintf
   #define stat            _stat
   #define putenv          _putenv
   #define t_func_t        void
   #define platform_new_thread(t_fun_ptr,param_ptr,id_ptr) \
      (*(id_ptr) = (void*)_beginthread(t_fun_ptr,0,(void*)(param_ptr)))
   #define platform_end_thread(x)

   #ifdef USING_OPENGL
      #include <GL/gl.h>
      #include <GL/glext.h>
      #define platform_ARB_load(n) wglGetProcAddress((LPCSTR)n)
   #endif

#else /* Use POSIX ... */

   #include <errno.h>
   #include <unistd.h>
   #include <pthread.h>
   #include <sys/time.h>
   #define Sleep(x)        usleep(1000 * (x))
   #define t_func_t        void*
   #define platform_new_thread(t_fun_ptr,param_ptr,id_ptr) \
      (!pthread_create((pthread_t*)(id_ptr),0,t_fun_ptr,(void*)(param_ptr)))
   #define platform_end_thread(x) \
      return (x)

   #ifdef USING_OPENGL
      #include <GL/gl.h>
      #include <GL/glx.h>
      #include <GL/glext.h>
      #define platform_ARB_load(n) glXGetProcAddressARB((const GLubyte*)n)
   #endif

#endif


double platform_delta_timer(void);

#endif /* __PLATFORM_H__ */
