#ifndef __RENDERER_API__
#define __RENDERER_API__
#include "types.h"
#include "matrix44.h"



#define RENDERER_NONE					0
#define RENDERER_DEFAULT				0
#define RENDERER_LEFT 					1
#define RENDERER_RIGHT 					2
#define RENDERER_BACK					3
#define RENDERER_FRONT					4

#define RENDERER_HORIZONTAL 			1
#define RENDERER_VERTICAL				2
#define RENDERER_CHECKER				3

#define RENDERER_COLOR					1
#define RENDERER_DEPTH					2

#define RENDERER_NOWRAP 				0
#define RENDERER_WRAP 					1

#define RENDERER_BYTE					1
#define RENDERER_FLOAT					2
#define RENDERER_INT						3

typedef struct {

	/* If list > 0 : use, else... */
	unsigned list;
	/* If vbo > 0 : use, else.. */
	unsigned vbo;

	/* Arrays: defined by type, width below. If 'vbo' is in effect, below
	values are offsets into the vbo array. */
	void *vtx;
	void *nrm;
	void *tex;
	void *clr;

	unsigned vtx_type;
	unsigned nrm_type;
	unsigned tex_type;
	unsigned clr_type;

	unsigned vtx_width;
	unsigned nrm_width;
	unsigned tex_width;
	unsigned clr_width;


}
render_call;

typedef void (*renderer_context_function)(void *reference, unsigned id);
typedef void (*renderer_generic_function)(void *reference);
typedef BITMAP_T renderer_bitmap;

/* Pre-usage, after rendering context has been created in display. */
void renderer_init(void);

/* Viewport */
void renderer_viewport(int x, int y, int width, int height);

/* Polygon and Line settings... */
void renderer_polygon_fill(void);
void renderer_polygon_wireframe(void);
void renderer_cull_back(void);
void renderer_cull_none(void);
void renderer_high_quality_lines(void);
void renderer_high_quality_polygons(void);

/* Framebuffer operations... */
void renderer_generate_stencil(int width, int height, unsigned type);
void renderer_use_buffer(unsigned buffer);
void renderer_use_stencil(unsigned stencil);
void renderer_use_colormask(unsigned colormask);

void renderer_clear_buffer(unsigned buffer);
void renderer_clear_color(float r, float g, float b, float a);

void renderer_use_depth_buffer(void);
void renderer_disable_depth_buffer(void);

/* Error messages... */
void renderer_print_error(void);

/* Compat.. remove eventually */
void renderer_use_built_in_lighting(float level);
void renderer_set_modelview(mat44 *m);
void renderer_set_projection(mat44 *m);


/* Texture mapping... */
unsigned renderer_map_texture(renderer_bitmap *bm, unsigned wrap, unsigned designate);

/* Buffer data mapping... */
void renderer_static_buffer_data(unsigned buffer,unsigned size,void *data);


/* Context management... */
/* All... */
void renderer_context_initialize(void);
void renderer_erase_registry(void);
/* Context groups.. */
unsigned renderer_generate_unique_id(void);
void renderer_set_context_group(unsigned id);
void renderer_activate_context_group(unsigned id);
void renderer_suppress_context_group(unsigned id);

/* Context registration callbacks... */
unsigned renderer_register_texture_file(const char *file);
unsigned renderer_register_texture_ptr(renderer_bitmap *bm);
unsigned renderer_register_alpha_map_ptr(renderer_bitmap *bm);

/* Get rid of... */
unsigned renderer_register_texture_function(void *owner, renderer_context_function ptr);
unsigned renderer_register_list_function(void *owner, renderer_context_function ptr);

/* ...replace with: */
unsigned renderer_register_vertex_buffer_function(void *owner, renderer_context_function ptr);
void renderer_register_generic(void *owner, renderer_generic_function begin, renderer_generic_function end);



/* Rendering... */
void renderer_draw_triangles(render_call *param);




#endif /* __RENDERER_API__ */