#ifdef __cplusplus
extern "C" {
#endif

#define USING_OPENGL
#include "platform.h"
#include "renderer_api.h"

#include "display.h"
#include "stage.h"
#include "ui_element.h"
#include "gluf.h"
#include "std_decor.h"
#include "display_controls.h"
#include "planck_scene.h"
#include "instruments_scene.h"
#include "credits_scene.h"
#include "root_scene.h"

#define  CONFIG_ID "MAIN_MENU"

enum {exit_all, go_planck, go_instr, go_credits, go_next, go_back, pop_menu};

typedef struct {

   IMPLEMENTS_SCENE_API

   ui_viewer      *viewer;
   ui_pager       *pgr;
   ui_balloon     *rndr;

   scene_api      *planck_scene;
   scene_api      *instruments_scene;
   scene_api      *credits_scene;

   unsigned show_planck_scene;
   unsigned show_instruments_scene;
   unsigned show_credits_scene;
   unsigned allow_exit;
   unsigned allow_right_click_menu;
   unsigned use_alternate_menu;

}
root_scene;


static void init(scene_api *object);
static void update(scene_api *object);
static void draw(scene_api *object);
static void enter(scene_api *object);
static void leave(scene_api *object);
static void erase(scene_api *object);
static void switch_board(void* obj,ui_update* info);
static void configuration(root_scene *rs);


static root_scene root_scene_proto = {
   &init,
   &update,
   &draw,
   &enter,
   &leave,
   &erase,
   0,
   1,
   0,0.001800,30.0,0.05,
   0,0,0,
   0,0,0,
   0,0,0,
   0,

   0,0,0,
   0,0,0,
   1,1,1,1,1,0
};

static void configuration(root_scene *rs)
{
   config_record *ret;

   rs->cl = create_config_loader(CONFIG_ID);
   aggregate_register_config(rs->ca,rs->cl);

   ret = config_bind(
      rs->cl,
      "show_planck_scene",
      CONFIG_UINT,
      sizeof(unsigned),
      &rs->show_planck_scene);
   config_doc(ret,
      "Controls inclusion of Planck launch item in main menu");
   ret = config_bind(
      rs->cl,
      "show_instruments_scene",
      CONFIG_UINT,
      sizeof(unsigned),
      &rs->show_instruments_scene);
   config_doc(ret,
      "Controls inclusion of Planck instrument explorer item in main menu");
   ret = config_bind(
      rs->cl,
      "show_credits_scene",
      CONFIG_UINT,
      sizeof(unsigned),
      &rs->show_credits_scene);
   config_doc(ret,
      "Controls inclusion of Credits item in main menu");
   ret = config_bind(
      rs->cl,
      "allow_exit",
      CONFIG_UINT,
      sizeof(unsigned),
      &rs->allow_exit);
   config_doc(ret,
      "Enables or disables the ability to exit the application");
   ret = config_bind(
      rs->cl,
      "allow_right_click_menu",
      CONFIG_UINT,
      sizeof(unsigned),
      &rs->allow_right_click_menu);
   config_doc(ret,
      "Controls access to display modes and audio controls");
   ret = config_bind(
      rs->cl,
      "use_alternate_menu",
      CONFIG_UINT,
      sizeof(unsigned),
      &rs->use_alternate_menu);
   config_doc(ret,
      "Enables an alternate icon based menu which has no exit function");

   aggregate_load_config(rs->ca,rs->cl);
}

scene_api *create_root_scene(unsigned designer, config_aggregate *ca)
{
   root_scene *create = (root_scene*)malloc(sizeof(root_scene));

   ui_icon *i;
   ui_container *m, *c, *am;
   ui_pager *bc;
   ui_button *bt;
   ui_element *e;
   ui_text *tx;

   if (create) {
      *create = root_scene_proto;
      create->ca = ca;

      configuration(create);

      create->planck_scene      = create_planck_scene(ca);
      create->instruments_scene = create_instruments_scene(designer,ca);
      create->credits_scene     = create_credits_scene(ca);

      create->context = renderer_generate_unique_id();
      renderer_set_context_group(create->context);

      create->viewer            = create_viewer(frame_state.overlay_area);

      /* Background image... */
      e = create_placard();
      decor_all_colors((ui_decor*)&e->look,1,1,1,1);
      decor_all_textures((ui_decor*)&e->look,renderer_register_texture_file(
         //"./data/models/solarSys/alt_sky.png"));
         "./data/images/maps/DSS2.jpg"));
      create->viewer->screen->add_item(create->viewer->screen,e);

      /* The main container... */
      m = create_container();
      m->owner  = create;
      m->on_use = &switch_board;
      m->id = pop_menu;
      m->button_mask = 48;
      create->viewer->screen->add_item(create->viewer->screen,m);

      /* Logos.. */
      e = create_placard();
      *((ui_decor*)&e->look) = std_static_decor;
      e->place(e,.875,.425,.11,.11);
      decor_all_textures((ui_decor*)&e->look,renderer_register_texture_file(
         "./data/images/misc/nasa_logo.png"));
      m->add_item(m,e);

      e = create_placard();
      *((ui_decor*)&e->look) = std_static_decor;
      e->place(e,.85,.3,.15,.15);
      decor_all_textures((ui_decor*)&e->look,renderer_register_texture_file(
         "./data/images/misc/esa_logo.png"));
      m->add_item(m,e);

      e = create_placard();
      *((ui_decor*)&e->look) = std_static_decor;
      e->place(e,.875,.25,.1,.1);
      decor_all_textures((ui_decor*)&e->look,renderer_register_texture_file(
         "./data/images/misc/jpl_logo.png"));
      m->add_item(m,e);

      e = create_placard();
      *((ui_decor*)&e->look) = std_static_decor;
      e->place(e,.875,.15,.1,.1);
      decor_all_textures((ui_decor*)&e->look,renderer_register_texture_file(
         "./data/images/misc/ucsb_logo.png"));
      m->add_item(m,e);

      e = create_placard();
      *((ui_decor*)&e->look) = std_static_decor;
      e->place(e,.875,.05,.1,.1);
      decor_all_textures((ui_decor*)&e->look,renderer_register_texture_file(
         "./data/images/misc/puc_logo.png"));
      m->add_item(m,e);

      /* Title... */
      i = create_custom_icon(.025);
      i->look = material_is(.2,.2,.2,.2,0);
      i->place(i,.0,.825,1,.1);
      m->add_item(m,i);

      tx = create_text(31,"The Planck Mission Simulation");
      tx->centered = 1;
      tx->font = ui_pretty_fnt;

      e = to_static_text(tx);
      e->place(e,0,.85,1,1);
      decor_all_colors((ui_decor*)&e->look,.2,.5,.2,.75);
      m->add_item(m,e);

      /* Icon pager... */
      i = create_custom_icon(.1);
      i->look = material_is(.2,.2,.2,.2,0);

      if (create->use_alternate_menu)
         i->place(i,.1875,.1,.625,.625);
      else
         i->place(i,.1875,.045,.625,.77);
      m->add_item(m,i);

      if (create->use_alternate_menu) {
         am = create_container();
         am->place(am,.1875,.1,.625,.625);
         m->add_item(m,am);
      }
      else {
         bc = create_pager();
         bc->place(bc,.2,.2,.6,.6);
         create->pgr = bc;
         m->add_item(m,bc);
      }

      /* The scene active icons (in the pager)... */
      if (create->show_planck_scene) {
         c = create_container();

         i = create_custom_icon(.1);
         *((ui_decor*)&i->look) = std_icon_decor;
         decor_all_textures((ui_decor*)&i->look,renderer_register_texture_file(
            "./data/images/misc/planck_scan_scene.png"));
         i->on_use = &switch_board;
         i->owner  = create;
         i->id = go_planck;
         c->add_item(c,i);

         tx = create_text(22,"Planck in Space");
         tx->centered = 1;
         tx->font = ui_pretty_fnt;
         e = to_static_text(tx);
         e->place(e,0,.1,1,1);
         decor_all_colors((ui_decor*)&e->look,.95,.95,.5,1);
         c->add_item(c,e);

         if (create->use_alternate_menu) {
            c->place(c,.075,.525,.4,.4);
            am->add_item(am,c);
         }
         else
            bc->add_item(bc,c);
      }

      if (create->show_instruments_scene) {
         c = create_container();

         i = create_custom_icon(.1);
         *((ui_decor*)&i->look) = std_icon_decor;
         decor_all_textures((ui_decor*)&i->look,renderer_register_texture_file(
            "./data/images/misc/instrument_scene.png"));
         i->on_use = &switch_board;
         i->owner  = create;
         i->id = go_instr;
         c->add_item(c,i);

         tx = create_text(22,"Planck Instruments");
         tx->centered = 1;
         tx->font = ui_pretty_fnt;
         e = to_static_text(tx);
         e->place(e,0,.1,1,1);
         decor_all_colors((ui_decor*)&e->look,.95,.95,.5,1);
         c->add_item(c,e);

         if (create->use_alternate_menu) {
            c->place(c,.525,.525,.4,.4);
            am->add_item(am,c);
         }
         else
            bc->add_item(bc,c);
      }

      if (create->show_credits_scene) {
         c = create_container();

         i = create_custom_icon(.1);
         *((ui_decor*)&i->look) = std_icon_decor;
         decor_all_textures((ui_decor*)&i->look,renderer_register_texture_file(
            "./data/images/misc/credit_scene.png"));
         i->on_use = &switch_board;
         i->owner  = create;
         i->id = go_credits;
         c->add_item(c,i);

         tx = create_text(22,"Acknowledgements");
         tx->centered = 1;
         tx->font = ui_pretty_fnt;
         e = to_static_text(tx);
         e->place(e,0,.1,1,1);
         decor_all_colors((ui_decor*)&e->look,.95,.95,.5,1);
         c->add_item(c,e);

         if (create->use_alternate_menu) {
            c->place(c,.075,.075,.4,.4);
            am->add_item(am,c);
         }
         else
            bc->add_item(bc,c);
      }

      /* Other buttons in the main container... */

      if (!create->use_alternate_menu) {

         if (create->allow_exit) {
            bt = create_button();
            bt->place(bt,.65,.06,.1,.1);
            *((ui_decor*)&bt->look) = std_rb_decor;
            bt->on_use = &switch_board;
            bt->owner  = create;
            bt->id = exit_all;
            m->add_item(m,bt);
            m->add_item(m,create_item_label(bt,"Exit"));
         }
         bt = create_button();
         bt->place(bt,.375,.06,.1,.1);
         *((ui_decor*)&bt->look) = std_bl_decor;
         bt->on_use = &switch_board;
         bt->owner  = create;
         bt->id = go_next;
         m->add_item(m,bt);
         m->add_item(m,create_item_label(bt,"Next"));

         bt = create_button();
         bt->place(bt,.25,.06,.1,.1);
         *((ui_decor*)&bt->look) = std_bl_decor;
         bt->on_use = &switch_board;
         bt->owner  = create;
         bt->id = go_back;
         m->add_item(m,bt);
         m->add_item(m,create_item_label(bt,"Back"));
      }

      /* The system options... */
      if (create->allow_right_click_menu) {
         create->rndr = create_display_stage_controls();
         create->viewer->screen->add_item(create->viewer->screen,create->rndr);
      }

      /* Beta tag
      tx = create_text(0,"Beta");
      tx->font = ui_pretty_fnt;

      e = to_static_text(tx);
      e->place(e,.2,0,1,1);
      e->orientation = 45;
      decor_all_colors((ui_decor*)&e->look,.3,.3,.3,.3);
      m->add_item(m,e); */
   }

   return (scene_api*)create;
}

static void erase(scene_api *object)
{
   root_scene *r = (root_scene*)object;

   if (!r)
      return;

   aggregate_save_config(r->ca,r->cl);

   if (r->viewer)
      r->viewer->erase(r->viewer);

   r->planck_scene->erase(r->planck_scene);
   r->instruments_scene->erase(r->instruments_scene);
   r->credits_scene->erase(r->credits_scene);

   free(r);
}

static void init(scene_api *object)
{
   root_scene *r = (root_scene*)object;

   if (!r)
      return;

   r->planck_scene->init(r->planck_scene);
   r->instruments_scene->init(r->instruments_scene);
   r->credits_scene->init(r->credits_scene);
}

static void switch_board(void* obj,ui_update* u)
{
   root_scene *r;
   ui_element *e;

   if (!obj)
      return;

   e = (ui_element*)obj;
   r = (root_scene*)e->owner;

   switch (e->id) {
      case exit_all:
         display_shutdown();
         break;
      case go_planck:
         stage_push_scene(r->planck_scene);
         break;
      case go_instr:
         stage_push_scene(r->instruments_scene);
         break;
      case go_credits:
         stage_push_scene(r->credits_scene);
         break;
      case go_next:
         r->pgr->next(r->pgr);
         break;
      case go_back:
         r->pgr->previous(r->pgr);
         break;
      case pop_menu:
         if (ctl_button(4,u->button) && r->rndr)
            r->rndr->grow(r->rndr);
         break;
   }
}

static void enter(scene_api *object)
{
   root_scene *r = (root_scene*)object;
}

static void leave(scene_api *object)
{
   root_scene *r = (root_scene*)object;
}

static void update(scene_api *object)
{
   root_scene *r = (root_scene*)object;

   if (!r)
      return;

   r->viewer->update(
      r->viewer,
      frame_state.pointer[0],
      frame_state.pointer[1],
      frame_state.dt,
      frame_state.button,
      frame_state.key);
}


static void draw(scene_api *object)
{
   root_scene *r = (root_scene*)object;

   if (!r)
      return;

   scene_api_gl_defaults();

   r->viewer->display(r->viewer, r->projection);
}

#ifdef __cplusplus
}
#endif
