#ifdef __cplusplus
extern "C" {
#endif


//#include "platform.h"
#include "renderer_api.h"
#include "display.h"
#include "scene_api.h"

static float illum_level = .5;


void scene_api_configure(scene_api *this_scene)
{
   if (!this_scene)
      return;

   if (display_set_clip_plane)
      display_set_clip_plane(this_scene->clipping_plane);
   if (display_set_eye_seperation)
      display_set_eye_seperation(this_scene->eye_seperation);
   if (display_set_focal_plane)
      display_set_focal_plane(this_scene->focal_plane);
}

void scene_api_enter(scene_api *this_scene)
{
   if (!this_scene)
      return;

   frame_state.rotation   = this_scene->rotation;
   frame_state.position   = this_scene->position;
   this_scene->projection = &frame_state.projection_matrix;

   scene_api_configure(this_scene);

   scene_api_set_brightness(this_scene->illumination);

   if (this_scene->enter)
      this_scene->enter(this_scene);

   if (this_scene->audio_level)
      this_scene->audio_level(this_scene,0);
}

void scene_api_leave(scene_api *this_scene)
{
   if (!this_scene)
      return;

   frame_state.rotation = 0;
   frame_state.position = 0;

   if (this_scene->audio_level)
      this_scene->audio_level(this_scene,0);

   if (this_scene->leave)
      this_scene->leave(this_scene);
}


void scene_api_set_brightness(float level)
{
   if (level > 1)
      level = 1;

   if (level < 0)
      level = 0;

   illum_level = level;
}

void scene_api_gl_defaults(void)
{
	renderer_use_depth_buffer();
	renderer_high_quality_polygons();
	renderer_high_quality_lines();

	//renderer_use_built_in_lighting(illum_level);

   renderer_clear_color(0,0,0,0);
}
#ifdef __cplusplus
}
#endif