#include "gl_extensions.h"
#include "ui_2d_shader.h"


static void enumerate(shader_program p);
static void execute(shader_program p);

ui_2d_shader_t ui_2d_shader = {

	INIT_MODULE_NAME("ui_2d_shader")

   INIT_VERTEX_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "attribute vec2 av2vertex;"
      "attribute vec2 av2texture;"

      "uniform mat4 um44mvp;"
      "uniform vec4 uv4color;"

      "varying vec4 vv4color;"
      "varying vec2 vv2texture;"

      "void main() {"
         "vv4color    = uv4color;"
         "vv2texture  = av2texture;"
         "gl_Position = um44mvp * vec4(av2vertex,0.0,1.0);"
      "}"

   INIT_FRAGMENT_SOURCE_BEGIN
      "#version 150\n"
      "precision highp float;"

      "uniform sampler2D texture1;"
      "uniform int have_texture;"
      "uniform int cull_back;"

      "varying vec4 vv4color;"
      "varying vec2 vv2texture;"

      "void main(void) {"
         "if (cull_back > 0 && !gl_FrontFacing)"
            "discard;"
         "vec4 tex = vec4(1.0);"
         "if (have_texture > 0)"
            "tex = texture2D(texture1,vv2texture);"
        "gl_FragColor =  vv4color * tex;"

      "}"

   INIT_ALLOCATED_SOURCES(0)
   INIT_STD_COMPILE_FUNCTION
   INIT_ENUMERATE_FUNCTION(&enumerate)
   INIT_EXECUTE_FUNCTION(&execute)
   INIT_SHADER_API_BLOCK_END
};

static GLint loc[7];

static void enumerate(shader_program p)
{
   ui_2d_shader_t *s = (ui_2d_shader_t*)p;

   if (!s || !glIsProgram(s->program))
      return;

   loc[0] = glGetUniformLocation(s->program,"um44mvp");
   loc[1] = glGetUniformLocation(s->program,"uv4color");
   loc[2] = glGetAttribLocation(s->program,"av2vertex");
   loc[3] = glGetAttribLocation(s->program,"av2texture");
   loc[4] = glGetUniformLocation(s->program,"have_texture");
   loc[5] = glGetUniformLocation(s->program,"texture1");
   loc[6] = glGetUniformLocation(s->program,"cull_back");

   //glDisableVertexAttribArray(loc[2]);
   //glDisableVertexAttribArray(loc[3]);
}

static void execute(shader_program p)
{
   ui_2d_shader_t *s = (ui_2d_shader_t*)p;

   if (!s)
      return;

   if (glIsProgram(s->program))
      glUseProgram(s->program);
   else
      return;

   /* Texture sampler: texture unit index 0. */
   glActiveTexture(GL_TEXTURE0);
   glUniform1i(loc[5],0);
   if (s->texture && glIsTexture(s->texture)) {
      glBindTexture(GL_TEXTURE_2D,s->texture);
      glUniform1i(loc[4],1);
   }
   else
      glUniform1i(loc[4],0);

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,s->vbo);

   if (s->cull_face)
      glUniform1i(loc[6],1);
    else
      glUniform1i(loc[6],0);


   if (s->depth_test)
      glEnable(GL_DEPTH_TEST);
   else
      glDisable(GL_DEPTH_TEST);

   glUniformMatrix4fv(loc[0],1,GL_FALSE,(float*)s->um44mvp);
   glUniform4fv(loc[1],1,s->uv4color);

   glVertexAttribPointer(loc[2],2,GL_FLOAT,GL_FALSE,0,s->av2vertex);
   glEnableVertexAttribArray(loc[2]);

   /* Weird stuff on ATI drivers: even if not used, texture attribute
   must be assinged to something and enabled, or nothing is displayed
   (shader crash?). */
   if (s->av2texture)
      glVertexAttribPointer(loc[3],2,GL_FLOAT,GL_FALSE,0,s->av2texture);
   else
      glVertexAttribPointer(loc[3],2,GL_FLOAT,GL_FALSE,0,s->av2vertex);

   glEnableVertexAttribArray(loc[3]);

   glDrawArrays(GL_TRIANGLES,0,s->count);

   if (glIsBuffer(s->vbo))
      glBindBuffer(GL_ARRAY_BUFFER,0);
   glDisableVertexAttribArray(loc[2]);
   glDisableVertexAttribArray(loc[3]);

   glUseProgram(0);
   glBindTexture(GL_TEXTURE_2D,0);
   glEnable(GL_DEPTH_TEST);
}
