#ifndef __UI_ELEMENT_H__
#define __UI_ELEMENT_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "types.h"
#include "gltf.h"
#include "matrix44.h"

extern gl_tex_font *ui_mono_fnt;
extern gl_tex_font *ui_pretty_fnt;
extern gl_tex_font *ui_pretty_bld_fnt;


typedef enum   {NoAct, Hover, Action}                          input_state;
//typedef struct {float x, y, w, h;}                             ui_space;
typedef vec4																	ui_space;
typedef SIMPLE_MAT_T               										ui_mat;
typedef struct {ui_mat look,hover,use;}                        ui_decor;
typedef struct {float px, py, dt; unsigned button; int key;}   ui_update;

typedef void (*ui_cursor_fn)(mat44*,mat44*);

/* ui_element */
typedef struct ui_element ui_element;
struct ui_element {
   #define IMPLEMENTS_UI_ELEMENT \
   ui_element *more;\
   ui_element *back;\
   ui_element *container;\
   void (*erase)(void *self);\
   void (*place)(void *self, float x, float y, float w, float h);\
   void (*on_display)(void *self, mat44 *projection, mat44 *modelview);\
   void (*update)(void *self,ui_update *info);\
   void (*on_hover)(void *self,ui_update *info);\
   void (*on_use)(void *self,ui_update *info);\
   void *owner;\
   unsigned id;\
   ui_space at;\
   float orientation;\
   float aspect;\
   input_state state;\
   unsigned button_mask;\
   int busy;\
   int sticky;\
   ui_mat look;\
   ui_mat hover_look;\
   ui_mat use_look;\
   vec2 *vtx;\
   vec2 *tex;\
   unsigned vtx_sz;\
   unsigned use_vtx;\
   unsigned disabled;\
   unsigned grab_focus;
   IMPLEMENTS_UI_ELEMENT
};

/* prototype ui_control; implements: ui_element */
typedef struct {
  IMPLEMENTS_UI_ELEMENT
  void* info; /* You MUST know what you are looking for. */
}
ui_control;

/* implements: ui_element */
typedef struct ui_container ui_container;
struct ui_container {
   #define AND_UI_CONTAINER \
   void (*add_item)(void *self,void *item);\
   void (*add_items)(void *self, int number_of_items, ...);\
   void (*to_list_head) (void *self,void *item);\
   void (*remove_item)(void *self, void *item);\
   void (*detach_item)(void *self, void *item);\
   void (*detach_all)(void *self);\
   ui_element *contains;\
   unsigned aspect_corrected;
   IMPLEMENTS_UI_ELEMENT
   AND_UI_CONTAINER
};

/* implements: ui_element, ui_container */
typedef struct ui_balloon ui_balloon;
struct ui_balloon {
   #define AND_UI_BALLOON \
   void (*grow)(ui_balloon *self);\
   void (*shrink)(ui_balloon *self);\
   void (*on_grow)(ui_balloon *self,ui_update *info);\
   void (*on_shrink)(ui_balloon *self,ui_update *info);\
   ui_space max;\
   float speed;\
   int change;\
   float presence;\
   unsigned expand_left;\
   unsigned expand_down;
   IMPLEMENTS_UI_ELEMENT
   AND_UI_CONTAINER
   AND_UI_BALLOON
};

/* implements: ui_element, ui_container */
typedef struct ui_pager ui_pager;
struct ui_pager {
   IMPLEMENTS_UI_ELEMENT
   AND_UI_CONTAINER
   void (*next)(ui_pager *self);
   void (*previous)(ui_pager *self);
   ui_element *show;
   ui_element *to;
   float speed;
   float change;
};

/* implements: ui_element */
typedef struct ui_text ui_text;
struct ui_text {
   #define AND_UI_TEXT \
   char* text;\
   gl_tex_font *font;\
   unsigned cpl;\
   unsigned justify_top;\
   unsigned centered;\
   unsigned force_monospace;
   IMPLEMENTS_UI_ELEMENT
   AND_UI_TEXT
};

typedef struct ui_text_value ui_text_value;
typedef void ui_text_updater(ui_text_value *object);
struct ui_text_value {
   IMPLEMENTS_UI_ELEMENT
   AND_UI_TEXT
   ui_text_updater *target;
   unsigned length;
   float time;
   float update_on;
};

/* implements: ui_element */
typedef struct ui_icon ui_icon;
struct ui_icon {
   IMPLEMENTS_UI_ELEMENT
   float edge;
};

typedef struct ui_element ui_placard;

/* implements: ui_element, ui_control */
typedef struct ui_dial ui_dial;
struct ui_dial {
   IMPLEMENTS_UI_ELEMENT
   float value;
   float span;
   float min;
   float speed;
};

/* implements: ui_element, ui_control */
typedef struct ui_selector_dial ui_selector_dial;
struct ui_selector_dial {
   IMPLEMENTS_UI_ELEMENT
   int value;
   float positions;
   float rotation;
   float dr;
   float speed;
};

/* implements: ui_element, ui_control */
typedef struct ui_slider ui_slider;
typedef struct ui_slider ui_progress_bar;
struct ui_slider {
   IMPLEMENTS_UI_ELEMENT
   float value;
   float span;
   float min;
   int detents;        /* n + 1 divisions of span... */
   float detent_width; /* as percentage of span... */
   float handle_width; /* as percentage of span... */
};

/* implements: ui_element, ui_control */
typedef struct ui_button ui_button;
struct ui_button {
   IMPLEMENTS_UI_ELEMENT
   int value;
   int value_down;
   int latches;
   int *chain;
   unsigned mtl_flag;
};

/* implements: ui_element, ui_control */
typedef struct ui_touch_pad ui_touch_pad;
struct ui_touch_pad {
   IMPLEMENTS_UI_ELEMENT
   float x;
   float y;
   float dead_zone;
};

/* implements: ui_element, ui_control */
typedef struct ui_marquee ui_marquee;
struct ui_marquee {
   IMPLEMENTS_UI_ELEMENT
   float position;
   float aperture;
   float scroll_speed; /* apertures per second */
};

/* adapter which brings together screen position, pointer, and control inputs. */
typedef struct ui_viewer ui_viewer;
struct ui_viewer {
   void (*erase)(void *self);
   void (*display)(void *self,mat44 *projection);
   void (*update)(void *self,float dx,float dy,float dt,unsigned button,int key);
   void (*pointer)(mat44 *projection, mat44 *modelview);
   ui_container *screen;

   /* *overlay format (in gl world coordinates/dimensions):
         [0] lower left x,
         [1] lower left y,
         [2] width,
         [3] height,
         [4] z depth

      Matches definition in renderer.h... */
   float *overlay;
   float aspect;
   float px;
   float py;

   /* action_mask default: button 4 or right mouse (value = 8). */
   unsigned action_mask;
};

/* implements: ui_element, ui_control */
typedef struct ui_key ui_key;
struct ui_key {
   IMPLEMENTS_UI_ELEMENT
   char value;
};

/* should be called once on start (loads fonts, etc.) and on termination... */
void              ui_initialize(void);
void              ui_end(void);

/* data manipulators, assignment tools. */
ui_space          in_space(float x, float y, float width, float height);
ui_mat            material_is(float r,float g,float b,float a, unsigned texture);
void              decor_all_textures(ui_decor *in,unsigned tex_id);
void              decor_all_colors(ui_decor *in, float r,float g,float b,float a);
void              decor_all_materials(ui_decor *in, float r,float g,float b,float a, unsigned texture);

/* ui object creation functions */
ui_container      *create_container(void);
ui_balloon        *create_balloon(void);
ui_pager          *create_pager(void);

ui_text           *create_text(unsigned chars_per_line, const char *text);
ui_text           *create_text_ptr(unsigned chars_per_line, char *text);
ui_text           *create_text_from_file(unsigned chars_per_line, const char *filename);
ui_text_value     *create_text_value(unsigned chars_per_line, unsigned length, ui_text_updater *target);


/* This function deletes the incoming text object and returns a generic element
which is composed of vertex and texture coordinate arrays which are normally
generated in a frame-by-frame basis in the mutable text object: these are
directly rendered, and so saves the overhead of text parsing... */
ui_element        *to_static_text(ui_text *text_object);

ui_element        *create_label(float x, float y, float size, const char *text);
ui_element        *create_item_label(void *item, const char *text);
ui_element        *create_item_lower_label(void *item, const char *text);
ui_element        *create_slider_labels(ui_slider *sl,unsigned number,...);

ui_icon           *create_icon(void);
ui_icon           *create_custom_icon(double edge_factor);
ui_placard        *create_placard(void);
ui_marquee        *create_marquee(void);
ui_progress_bar   *create_progress_bar(void);

ui_dial           *create_dial(void);
ui_selector_dial  *create_selector(void);
ui_slider         *create_slider(void);
ui_button         *create_button(void);
ui_touch_pad      *create_touch_pad(void);

ui_viewer         *create_viewer(float *overlay);

ui_key            *create_key(void);

void              ui_basic_cursor(mat44 *projection, mat44 *modelview);
void              ui_busy_cursor(mat44 *projection, mat44 *modelview);

/* If manual deletion of objects is desired (not recommended), call the
built in "erase" function of the object e.g. ((ui_list*)item)->erase(item). */

#ifdef __cplusplus
}
#endif

#endif /* __UI_ELEMENT_H__ */